# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_trace_protos/trace.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bpw_trace_protos/trace.proto\x12\x08pw.trace\"\'\n\x11SnapshotTraceInfo\x12\x12\n\ntrace_data\x18\x15 \x01(\x0c\x42\x17\n\x0epw.trace.protoB\x05Traceb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_trace_protos.trace_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\016pw.trace.protoB\005Trace'
  _globals['_SNAPSHOTTRACEINFO']._serialized_start=41
  _globals['_SNAPSHOTTRACEINFO']._serialized_end=80
# @@protoc_insertion_point(module_scope)



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
# pylint: disable=line-too-long
"""Serializes an Environment into files in a way GitHub Actions understands.

See also:

* https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#setting-an-environment-variable
* https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#adding-a-system-path
"""
# pylint: enable=line-too-long

from __future__ import print_function

import contextlib
import os

# Disable Python 2-related warnings since this file must be Python 2
# compatible.
# pylint: disable=super-with-arguments, useless-object-inheritance


class GitHubVisitor(object):
    """Serializes an Environment into files GitHub Actions understands."""

    def __init__(self, *args, **kwargs):
        super(GitHubVisitor, self).__init__(*args, **kwargs)
        self._replacements = ()
        self._github_env = None
        self._github_path = None
        self._log = None

    def serialize(self, env, root):
        """Write a shell file based on the given environment.

        Args:
            env (environment.Environment): Environment variables to use.
            outs (file): Shell file to write.
        """
        try:
            self._replacements = tuple(
                (key, env.get(key) if value is None else value)
                for key, value in env.replacements
            )

            with contextlib.ExitStack() as stack:
                github_env = os.environ.get('GITHUB_ENV')
                mode = 'a'
                if not github_env:
                    github_env = os.path.join(root, 'github_env.log')
                    mode = 'w'
                self._github_env = stack.enter_context(open(github_env, mode))

                github_path = os.environ.get('GITHUB_PATH')
                mode = 'a'
                if not github_path:
                    github_path = os.path.join(root, 'github_path.log')
                    mode = 'w'
                self._github_path = stack.enter_context(open(github_path, mode))

                self._log = stack.enter_context(
                    open(os.path.join(root, 'github.log'), 'w')
                )

                env.accept(self)

        finally:
            self._replacements = ()
            self._github_env = None
            self._github_path = None
            self._log = None

    def visit_set(self, set):  # pylint: disable=redefined-builtin
        if '\n' in set.value:
            eof = '__EOF__'
            assert '\n{}\n'.format(eof) not in set.value
            print(
                '{name}=<<{eof}\n{value}\n{eof}'.format(
                    name=set.name,
                    value=set.value,
                    eof=eof,
                ),
                file=self._github_env,
            )
        else:
            print(
                '{name}={value}'.format(name=set.name, value=set.value),
                file=self._github_env,
            )
        print(
            'setting {name!r} = {value!r}'.format(
                name=set.name, value=set.value
            ),
            file=self._log,
        )

    def visit_clear(self, clear):
        print('{}='.format(clear.name), file=self._github_env)
        print('setting {!r} = ""'.format(clear.name), file=self._log)

    def visit_prepend(self, prepend):
        if prepend.name == 'PATH':
            print(prepend.value, file=self._github_path)
            print('adding {!r} to PATH'.format(prepend.value), file=self._log)
        else:
            print(
                'unsupported prepend: {name!r} += {value!r}'.format(
                    name=prepend.name, value=prepend.value
                ),
                file=self._log,
            )

    def visit_append(self, append):
        print(
            'unsupported append: {name!r} += {value!r}'.format(
                name=append.name, value=append.value
            ),
            file=self._log,
        )

    def visit_remove(self, remove):
        pass

    def visit_echo(self, echo):
        pass

    def visit_comment(self, comment):
        pass

    def visit_command(self, command):
        pass

    def visit_doctor(self, doctor):
        pass

    def visit_blank_line(self, blank_line):
        pass

    def visit_function(self, function):
        pass



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Preconfigured checks for Python code.

These checks assume that they are running in a preconfigured Python environment.
"""

import difflib
import json
import logging
from pathlib import Path
import platform
import re
import shutil
import sys
from tempfile import TemporaryDirectory
import venv

from pw_cli.diff import colorize_diff_line
from pw_env_setup import python_packages

from pw_presubmit.presubmit import (
    call,
    Check,
    filter_paths,
)
from pw_presubmit.git_repo import LoggingGitRepo
from pw_presubmit.presubmit_context import (
    PresubmitContext,
    PresubmitFailure,
)
from pw_presubmit import build
from pw_presubmit.tools import log_run

_LOG = logging.getLogger(__name__)

_PYTHON_EXTENSIONS = ('.py', '.gn', '.gni')

_PYTHON_PACKAGE_EXTENSIONS = (
    'setup.cfg',
    'constraint.list',
    'requirements.txt',
)

_PYTHON_IS_3_9_OR_HIGHER = sys.version_info >= (
    3,
    9,
)


@filter_paths(endswith=_PYTHON_EXTENSIONS)
def gn_python_check(ctx: PresubmitContext):
    build.gn_gen(ctx)
    build.ninja(ctx, 'python.tests', 'python.lint')


def _transform_lcov_file_paths(lcov_file: Path, repo_root: Path) -> str:
    """Modify file paths in an lcov file to be relative to the repo root.

    See `man geninfo` for info on the lcov format."""

    lcov_input = lcov_file.read_text()
    lcov_output = ''

    if not _PYTHON_IS_3_9_OR_HIGHER:
        return lcov_input

    for line in lcov_input.splitlines():
        if not line.startswith('SF:'):
            lcov_output += line + '\n'
            continue

        # Get the file path after SF:
        file_string = line[3:].rstrip()
        source_file_path = Path(file_string)

        # Attempt to map a generated Python package source file to the root
        # source tree.
        # pylint: disable=no-member
        if not source_file_path.is_relative_to(
            repo_root  # type: ignore[attr-defined]
        ):
            # pylint: enable=no-member
            source_file_path = repo_root / str(source_file_path).replace(
                'python/gen/', ''
            ).replace('py.generated_python_package/', '')

        # If mapping fails don't modify this line.
        # pylint: disable=no-member
        if not source_file_path.is_relative_to(
            repo_root  # type: ignore[attr-defined]
        ):
            # pylint: enable=no-member
            lcov_output += line + '\n'
            continue

        source_file_path = source_file_path.relative_to(repo_root)
        lcov_output += f'SF:{source_file_path}\n'

    return lcov_output


@filter_paths(endswith=_PYTHON_EXTENSIONS)
def gn_python_test_coverage(ctx: PresubmitContext):
    """Run Python tests with coverage and create reports."""
    build.gn_gen(ctx, pw_build_PYTHON_TEST_COVERAGE=True)
    build.ninja(ctx, 'python.tests')

    # Find coverage data files
    coverage_data_files = list(ctx.output_dir.glob('**/*.coverage'))
    if not coverage_data_files:
        return

    # Merge coverage data files to out/.coverage
    call(
        'coverage',
        'combine',
        # Leave existing coverage files in place; by default they are deleted.
        '--keep',
        *coverage_data_files,
        cwd=ctx.output_dir,
    )
    combined_data_file = ctx.output_dir / '.coverage'
    _LOG.info('Coverage data saved to: %s', combined_data_file.resolve())

    # Always ignore generated proto python and setup.py files.
    coverage_omit_patterns = '--omit=*_pb2.py,*/setup.py'

    # Output coverage percentage summary to the terminal of changed files.
    changed_python_files = list(
        str(p) for p in ctx.paths if str(p).endswith('.py')
    )
    report_args = [
        'coverage',
        'report',
        '--ignore-errors',
        coverage_omit_patterns,
    ]
    report_args += changed_python_files
    log_run(report_args, check=False, cwd=ctx.output_dir)

    # Generate a json report
    call('coverage', 'lcov', coverage_omit_patterns, cwd=ctx.output_dir)
    lcov_data_file = ctx.output_dir / 'coverage.lcov'
    lcov_data_file.write_text(
        _transform_lcov_file_paths(lcov_data_file, repo_root=ctx.root)
    )
    _LOG.info('Coverage lcov saved to: %s', lcov_data_file.resolve())

    # Generate an html report
    call('coverage', 'html', coverage_omit_patterns, cwd=ctx.output_dir)
    html_report = ctx.output_dir / 'htmlcov' / 'index.html'
    _LOG.info('Coverage html report saved to: %s', html_report.resolve())


@filter_paths(endswith=_PYTHON_PACKAGE_EXTENSIONS)
def vendor_python_wheels(ctx: PresubmitContext) -> None:
    """Download Python packages locally for the current platform."""
    build.gn_gen(ctx)
    build.ninja(ctx, 'pip_vendor_wheels')

    download_log = (
        ctx.output_dir
        / 'python/gen/pw_env_setup/pigweed_build_venv.vendor_wheels'
        / 'pip_download_log.txt'
    )
    _LOG.info('Python package download log: %s', download_log)

    wheel_output = (
        ctx.output_dir
        / 'python/gen/pw_env_setup'
        / 'pigweed_build_venv.vendor_wheels/wheels/'
    )
    wheel_destination = ctx.output_dir / 'python_wheels'
    shutil.rmtree(wheel_destination, ignore_errors=True)
    shutil.copytree(wheel_output, wheel_destination, dirs_exist_ok=True)

    _LOG.info('Python packages downloaded to: %s', wheel_destination)


def _version_bump_setup_cfg(
    repo_root: Path, setup_cfg: Path, modify_setup_cfg: bool = True
) -> str:
    """Increment the version patch number of a setup.cfg

    Skips modifying if there are modifications since origin/main.

    Returns:
       The version number as a string
    """
    repo = LoggingGitRepo(repo_root)
    setup_cfg = repo_root / setup_cfg

    # Grab the current version string.
    _LOG.info('Checking the version patch number in: %s', setup_cfg)
    setup_cfg_text = setup_cfg.read_text()
    version_regex = (
        r'^version = (?P<version>'
        r'(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)'
        r')$'
    )
    match = re.search(version_regex, setup_cfg_text, re.MULTILINE)
    if not match:
        raise PresubmitFailure(
            f'Unable to find "version = x.x.x" line in {setup_cfg}'
        )
    _LOG.info('Found: %s', match[0])
    version_number = match['version']

    # Skip modifying the version if it is different compared to origin/main.
    modified_files = repo.list_files(commit='origin/main')
    if modify_setup_cfg and setup_cfg in modified_files:
        # Don't update the file
        modify_setup_cfg = False
        _LOG.warning(
            '%s is already modified, skipping version update.', setup_cfg
        )

    if modify_setup_cfg:
        # Increment the patch number.
        try:
            patch_number = int(match['patch']) + 1
        except ValueError as err:
            raise PresubmitFailure(
                f"Unable to increment patch number: '{match['patch']}' "
                f"for version line: '{match[0]}'"
            ) from err

        version_number = f"{match['major']}.{match['minor']}.{patch_number}"
        new_line = f'version = {version_number}'
        new_text = re.sub(
            version_regex, new_line, setup_cfg_text, count=1, flags=re.MULTILINE
        )

        # Print the diff
        setup_cfg_diff = list(
            difflib.unified_diff(
                setup_cfg_text.splitlines(),
                new_text.splitlines(),
                fromfile=str(setup_cfg) + ' (original)',
                tofile=str(setup_cfg) + ' (updated)',
                lineterm='',
                n=1,
            )
        )
        if setup_cfg_diff:
            for line in setup_cfg_diff:
                print(colorize_diff_line(line))

        # Update the file.
        setup_cfg.write_text(new_text, encoding='utf-8')

    return version_number


def version_bump_pigweed_pypi_distribution(ctx: PresubmitContext) -> None:
    _version_bump_setup_cfg(
        repo_root=ctx.root,
        setup_cfg=ctx.root / 'pw_env_setup/pypi_common_setup.cfg',
    )


def upload_pigweed_pypi_distribution(
    ctx: PresubmitContext,
) -> None:
    """Version bump the pigweed pypi distribution and upload to pypi.org.

    This requires an API token to be setup for the current user. See also:
    https://pypi.org/help/#apitoken
    https://packaging.python.org/en/latest/guides/distributing-packages-using-setuptools/#create-an-account
    """
    version_number = _version_bump_setup_cfg(
        repo_root=ctx.root,
        setup_cfg=ctx.root / 'pw_env_setup/pypi_common_setup.cfg',
        modify_setup_cfg=False,
    )

    _LOG.info('Cleaning any existing build artifacts.')
    build.gn_gen(ctx)

    dist_output_path = (
        ctx.output_dir
        / 'python/obj/pw_env_setup/pypi_pigweed_python_source_tree'
    )

    # Always remove any existing build artifacts before generating a
    # new distribution. 'python -m build' leaves a 'dist' directory without
    # cleaning up.
    shutil.rmtree(dist_output_path, ignore_errors=True)
    build.ninja(ctx, 'pigweed_pypi_distribution', '-t', 'clean')

    # Generate the distribution
    _LOG.info('Running the ninja build.')
    build.ninja(ctx, 'pigweed_pypi_distribution')

    # Check the output is in the right place.
    if any(
        not (dist_output_path / f).is_file() for f in ['README.md', 'LICENSE']
    ):
        raise PresubmitFailure(
            f'Missing pypi distribution files in: {dist_output_path}'
        )

    # Create a new venv for building and uploading.
    venv_path = ctx.output_dir / 'pypi_upload_venv'
    _LOG.info('Creating venv for uploading in: %s', venv_path)
    shutil.rmtree(venv_path, ignore_errors=True)
    venv.create(venv_path, symlinks=True, with_pip=True)
    py_bin = venv_path / 'bin/python'

    # Install upload tools.
    _LOG.info('Running: pip install --upgrade pip %s', venv_path)
    log_run(
        [py_bin, '-m', 'pip', 'install', '--upgrade', 'pip'],
        check=True,
        cwd=ctx.output_dir,
    )
    _LOG.info('Running: pip install --upgrade build twine %s', venv_path)
    log_run(
        [py_bin, '-m', 'pip', 'install', '--upgrade', 'build', 'twine'],
        check=True,
        cwd=ctx.output_dir,
    )

    # Create upload artifacts
    _LOG.info('Running: python -m build')
    log_run([py_bin, '-m', 'build'], check=True, cwd=dist_output_path)

    dist_path = dist_output_path / 'dist'
    upload_files = list(dist_path.glob('*'))
    expected_files = [
        dist_path / f'pigweed-{version_number}.tar.gz',
        dist_path / f'pigweed-{version_number}-py3-none-any.whl',
    ]
    if upload_files != expected_files:
        raise PresubmitFailure(
            'Unexpected dist files found for upload. Skipping upload.\n'
            f'Found:\n {upload_files}\n'
            f'Expected:\n {expected_files}\n'
        )

    # Upload to pypi.org
    upload_args = [py_bin, '-m', 'twine', 'upload']
    upload_args.extend(expected_files)
    log_run(upload_args, check=True, cwd=dist_output_path)


def _generate_constraint_with_hashes(
    ctx: PresubmitContext, input_file: Path, output_file: Path
) -> None:
    assert input_file.is_file()

    call(
        "pip-compile",
        input_file,
        "--generate-hashes",
        "--reuse-hashes",
        "--resolver=backtracking",
        "--strip-extras",
        # Force pinning pip and setuptools
        "--allow-unsafe",
        "-o",
        output_file,
    )

    # Remove absolute paths from comments
    output_text = output_file.read_text()
    output_text = output_text.replace(str(ctx.output_dir), '')
    output_text = output_text.replace(str(ctx.root), '')
    output_text = output_text.replace(str(output_file.parent), '')

    final_output_text = ''
    for line in output_text.splitlines(keepends=True):
        # Remove --find-links lines
        if line.startswith('--find-links'):
            continue
        # Remove blank lines
        if line == '\n':
            continue
        final_output_text += line

    output_file.write_text(final_output_text)


def _update_upstream_python_constraints(
    ctx: PresubmitContext,
    update_files: bool = False,
) -> None:
    """Regenerate platform specific Python constraint files with hashes."""
    with TemporaryDirectory() as tmpdirname:
        out_dir = Path(tmpdirname)
        build.gn_gen(
            ctx,
            pw_build_PIP_REQUIREMENTS=[],
            # Use the constraint file without hashes as the input. This is where
            # new packages are added by developers.
            pw_build_PIP_CONSTRAINTS=[
                '//pw_env_setup/py/pw_env_setup/virtualenv_setup/'
                'constraint.list',
            ],
            # This should always be set to false when regenrating constraints.
            pw_build_PYTHON_PIP_INSTALL_REQUIRE_HASHES=False,
        )
        build.ninja(ctx, 'pip_constraint_update')

        # Either: darwin, linux or windows
        platform_name = platform.system().lower()

        constraint_hashes_filename = f'constraint_hashes_{platform_name}.list'
        constraint_hashes_original = (
            ctx.root
            / 'pw_env_setup/py/pw_env_setup/virtualenv_setup'
            / constraint_hashes_filename
        )
        constraint_hashes_tmp_out = out_dir / constraint_hashes_filename
        _generate_constraint_with_hashes(
            ctx,
            input_file=(
                ctx.output_dir
                / 'python/gen/pw_env_setup/pigweed_build_venv'
                / 'compiled_requirements.txt'
            ),
            output_file=constraint_hashes_tmp_out,
        )

        build.gn_gen(
            ctx,
            # This should always be set to false when regenrating constraints.
            pw_build_PYTHON_PIP_INSTALL_REQUIRE_HASHES=False,
        )
        build.ninja(ctx, 'pip_constraint_update')

        upstream_requirements_lock_filename = (
            f'upstream_requirements_{platform_name}_lock.txt'
        )
        upstream_requirements_lock_original = (
            ctx.root
            / 'pw_env_setup/py/pw_env_setup/virtualenv_setup'
            / upstream_requirements_lock_filename
        )
        upstream_requirements_lock_tmp_out = (
            out_dir / upstream_requirements_lock_filename
        )
        _generate_constraint_with_hashes(
            ctx,
            input_file=(
                ctx.output_dir
                / 'python/gen/pw_env_setup/pigweed_build_venv'
                / 'compiled_requirements.txt'
            ),
            output_file=upstream_requirements_lock_tmp_out,
        )

        if update_files:
            constraint_hashes_original.write_text(
                constraint_hashes_tmp_out.read_text()
            )
            _LOG.info('Updated: %s', constraint_hashes_original)
            upstream_requirements_lock_original.write_text(
                upstream_requirements_lock_tmp_out.read_text()
            )
            _LOG.info('Updated: %s', upstream_requirements_lock_original)
            return

        # Make a diff of required changes
        constraint_hashes_diff = list(
            difflib.unified_diff(
                constraint_hashes_original.read_text(
                    'utf-8', errors='replace'
                ).splitlines(),
                constraint_hashes_tmp_out.read_text(
                    'utf-8', errors='replace'
                ).splitlines(),
                fromfile=str(constraint_hashes_original) + ' (original)',
                tofile=str(constraint_hashes_original) + ' (updated)',
                lineterm='',
                n=1,
            )
        )
        upstream_requirements_lock_diff = list(
            difflib.unified_diff(
                upstream_requirements_lock_original.read_text(
                    'utf-8', errors='replace'
                ).splitlines(),
                upstream_requirements_lock_tmp_out.read_text(
                    'utf-8', errors='replace'
                ).splitlines(),
                fromfile=str(upstream_requirements_lock_original)
                + ' (original)',
                tofile=str(upstream_requirements_lock_original) + ' (updated)',
                lineterm='',
                n=1,
            )
        )
        if constraint_hashes_diff:
            for line in constraint_hashes_diff:
                print(colorize_diff_line(line))
        if upstream_requirements_lock_diff:
            for line in upstream_requirements_lock_diff:
                print(colorize_diff_line(line))
        if constraint_hashes_diff or upstream_requirements_lock_diff:
            raise PresubmitFailure(
                'Please run:\n'
                '\n'
                '  pw presubmit --step update_upstream_python_constraints'
            )


@filter_paths(endswith=_PYTHON_PACKAGE_EXTENSIONS)
def check_upstream_python_constraints(ctx: PresubmitContext) -> None:
    _update_upstream_python_constraints(ctx, update_files=False)


@filter_paths(endswith=_PYTHON_PACKAGE_EXTENSIONS)
def update_upstream_python_constraints(ctx: PresubmitContext) -> None:
    _update_upstream_python_constraints(ctx, update_files=True)


@filter_paths(endswith=_PYTHON_EXTENSIONS + ('.pylintrc',))
def gn_python_lint(ctx: PresubmitContext) -> None:
    build.gn_gen(ctx)
    build.ninja(ctx, 'python.lint')


@Check
def check_python_versions(ctx: PresubmitContext):
    """Checks that the list of installed packages is as expected."""

    build.gn_gen(ctx)
    constraint_file: str | None = None
    requirement_file: str | None = None
    try:
        for arg in build.get_gn_args(ctx.output_dir):
            if arg['name'] == 'pw_build_PIP_CONSTRAINTS':
                constraint_file = json.loads(arg['current']['value'])[0].strip(
                    '/'
                )
            if arg['name'] == 'pw_build_PIP_REQUIREMENTS':
                requirement_file = json.loads(arg['current']['value'])[0].strip(
                    '/'
                )
    except json.JSONDecodeError:
        _LOG.warning('failed to parse GN args json')
        return

    if not constraint_file:
        _LOG.warning('could not find pw_build_PIP_CONSTRAINTS GN arg')
        return
    ignored_requirements_arg = None
    if requirement_file:
        ignored_requirements_arg = [(ctx.root / requirement_file)]

    if (
        python_packages.diff(
            expected=(ctx.root / constraint_file),
            ignore_requirements_file=ignored_requirements_arg,
        )
        != 0
    ):
        raise PresubmitFailure



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Crate a venv."""

import argparse
import os
import pathlib
import platform
import shutil
import stat
import sys
import venv


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--depfile',
        type=pathlib.Path,
        required=True,
        help='Path at which a depfile should be written.',
    )
    parser.add_argument(
        '--destination-dir',
        type=pathlib.Path,
        required=True,
        help='Path to venv directory.',
    )
    parser.add_argument(
        '--stampfile',
        type=pathlib.Path,
        required=True,
        help="Path to this target's stamp file.",
    )
    return parser.parse_args()


def _rm_dir(path_to_delete: pathlib.Path) -> None:
    """Delete a directory recursively.

    On Windows if a file can't be deleted, mark it as writable then delete.
    """

    def make_writable_and_delete(_func, path, _exc_info):
        os.chmod(path, stat.S_IWRITE)
        os.unlink(path)

    on_rm_error = None
    if platform.system() == 'Windows':
        on_rm_error = make_writable_and_delete
    shutil.rmtree(path_to_delete, onerror=on_rm_error)


def main(
    depfile: pathlib.Path,
    destination_dir: pathlib.Path,
    stampfile: pathlib.Path,
) -> None:
    # Create the virtualenv.
    if destination_dir.exists():
        _rm_dir(destination_dir)
    venv.create(destination_dir, symlinks=True, with_pip=True)

    # Write out the depfile, making sure the Python path is
    # relative to the outdir so that this doesn't add user-specific
    # info to the build.
    out_dir_path = pathlib.Path(os.getcwd()).resolve()
    python_path = pathlib.Path(sys.executable).resolve()
    rel_python_path = os.path.relpath(python_path, start=out_dir_path)
    depfile.write_text(f'{stampfile}: \\\n  {rel_python_path}')


if __name__ == '__main__':
    main(**vars(_parse_args()))



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Check the formatting of TODOs."""

import logging
from pathlib import Path
import re
from typing import Iterable, Pattern, Sequence

from pw_presubmit import presubmit_context
from pw_presubmit.presubmit import filter_paths
from pw_presubmit.presubmit_context import PresubmitContext

_LOG: logging.Logger = logging.getLogger(__name__)

EXCLUDE: Sequence[str] = (
    # Metadata
    r'^docker/tag$',
    r'\byarn.lock$',
    # Data files
    r'\.bin$',
    r'\.csv$',
    r'\.elf$',
    r'\.gif$',
    r'\.jpg$',
    r'\.json$',
    r'\.png$',
    r'\.svg$',
    r'\.xml$',
)

# todo-check: disable
# pylint: disable=line-too-long
BUGS_ONLY = re.compile(
    r'(?:\bTODO\(b/\d+(?:, ?b/\d+)*\).*\w)|'
    r'(?:\bTODO: b/\d+(?:, ?b/\d+)* - )|'
    r'(?:\bTODO: https://issues\.(?:pigweed|fuchsia)\.dev/(?:issues/)?\d+ - )|'
    r'(?:\bTODO: https://issues\.chromium\.org/(?:issues/)?\d+ - )|'
    r'(?:\bTODO: https://(?:(?:pw|fx)bug\.dev|crbug\.com)/\d+ - )|'
    r'(?:\bTODO: (?:(?:pw|fx)bug\.dev|crbug\.com)/\d+ - )|'
    r'(?:\bTODO: <(?:(?:pw|fx)bug\.dev|crbug\.com)/\d+> - )|'
    r'(?:\bTODO: https://github\.com/bazelbuild/[a-z][-_a-z0-9]*/issues/\d+[ ]-[ ])'
)
BUGS_OR_USERNAMES = re.compile(
    r"""
(?:  # Legacy style.
    \bTODO\(
        (?:b/\d+|[a-z]+)  # Username or bug.
        (?:,[ ]?(?:b/\d+|[a-z]+))*  # Additional usernames or bugs.
    \)
.*\w  # Explanation.
)|
(?:  # New style.
    \bTODO:[ ]
    (?:
        b/\d+|  # Bug.
        https://(?:(?:pw|fx)bug\.dev|crbug\.com)(?:/issues)?/\d+| # Short URL
        (?:(?:pw|fx)bug\.dev|crbug\.com)(?:/issues)?/\d+| # Even shorter URL
        <(?:(?:pw|fx)bug\.dev|crbug\.com)/\d+>| # Markdown compatible
        https://issues\.(?:pigweed|fuchsia)\.dev(?:/issues)?/\d+| # Fully qualified bug for rustdoc
        https://issues.chromium.org/issues/\d+| # Fully qualified bug for rustdoc
        # Username@ with optional domain.
        [a-z]+@(?:[a-z][-a-z0-9]*(?:\.[a-z][-a-z0-9]*)+)?
    )
    (?:,[ ]?  # Additional.
        (?:
            b/\d+|  # Bug.
            # Username@ with optional domain.
            [a-z]+@(?:[a-z][-a-z0-9]*(?:\.[a-z][-a-z0-9]*)+)?
        )
    )*
[ ]-[ ].*\w  # Explanation.
)|
(?:  # Fuchsia style.
    \bTODO\(
        (?:fxbug\.dev/\d+|[a-z]+)  # Username or bug.
        (?:,[ ]?(?:fxbug\.dev/\d+|[a-z]+))*  # Additional usernames or bugs.
    \)
.*\w  # Explanation.
)|
(?:  # Bazel GitHub issues. No usernames allowed.
    \bTODO:[ ]
    (?:
        https://github\.com/bazelbuild/[a-z][-_a-z0-9]*/issues/\d+
    )
[ ]-[ ].*\w  # Explanation.
)
    """,
    re.VERBOSE,
)
# pylint: enable=line-too-long

_TODO_OR_FIXME = re.compile(r'(\bTODO\b)|(\bFIXME\b)')
# todo-check: enable

# If seen, ignore this line and the next.
_IGNORE = 'todo-check: ignore'

# Ignore a whole section. Please do not change the order of these lines.
_DISABLE = 'todo-check: disable'
_ENABLE = 'todo-check: enable'


def _process_file(ctx: PresubmitContext, todo_pattern: re.Pattern, path: Path):
    with path.open() as ins:
        _LOG.debug('Evaluating path %s', path)
        enabled = True
        prev = ''

        try:
            summary: list[str] = []
            for i, line in enumerate(ins, 1):
                if _DISABLE in line:
                    enabled = False
                elif _ENABLE in line:
                    enabled = True

                if not enabled or _IGNORE in line or _IGNORE in prev:
                    prev = line
                    continue

                if _TODO_OR_FIXME.search(line):
                    if not todo_pattern.search(line):
                        # todo-check: ignore
                        ctx.fail(f'Bad TODO on line {i}:', path)
                        ctx.fail(f'    {line.strip()}')
                        ctx.fail('Prefer this format in new code:')
                        # todo-check: ignore
                        ctx.fail(
                            '    TODO: https://pwbug.dev/12345 - More context.'
                        )
                        summary.append(f'{i}:{line.strip()}')

                prev = line

            return summary

        except UnicodeDecodeError:
            # File is not text, like a gif.
            _LOG.debug('File %s is not a text file', path)
            return []


def create(
    todo_pattern: re.Pattern = BUGS_ONLY,
    exclude: Iterable[Pattern[str] | str] = EXCLUDE,
):
    """Create a todo_check presubmit step that uses the given pattern."""

    @filter_paths(exclude=exclude)
    def todo_check(ctx: PresubmitContext):
        """Check that TODO lines are valid."""  # todo-check: ignore
        ctx.paths = presubmit_context.apply_exclusions(ctx)
        summary: dict[Path, list[str]] = {}
        for path in ctx.paths:
            if file_summary := _process_file(ctx, todo_pattern, path):
                summary[path] = file_summary

        if summary:
            with ctx.failure_summary_log.open('w') as outs:
                for path, lines in summary.items():
                    print('====', path.relative_to(ctx.root), file=outs)
                    for line in lines:
                        print(line, file=outs)

    return todo_check



# Copyright 2019 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Renders HTML documentation using Sphinx."""

# TODO(frolv): Figure out a solution for installing all library dependencies
# to run Sphinx and build RTD docs.

import argparse
import collections
import json
import os
import shutil
import subprocess
import sys

from pathlib import Path

SCRIPT_HEADER: str = '''
██████╗ ██╗ ██████╗ ██╗    ██╗███████╗███████╗██████╗     ██████╗  ██████╗  ██████╗███████╗
██╔══██╗██║██╔════╝ ██║    ██║██╔════╝██╔════╝██╔══██╗    ██╔══██╗██╔═══██╗██╔════╝██╔════╝
██████╔╝██║██║  ███╗██║ █╗ ██║█████╗  █████╗  ██║  ██║    ██║  ██║██║   ██║██║     ███████╗
██╔═══╝ ██║██║   ██║██║███╗██║██╔══╝  ██╔══╝  ██║  ██║    ██║  ██║██║   ██║██║     ╚════██║
██║     ██║╚██████╔╝╚███╔███╔╝███████╗███████╗██████╔╝    ██████╔╝╚██████╔╝╚██████╗███████║
╚═╝     ╚═╝ ╚═════╝  ╚══╝╚══╝ ╚══════╝╚══════╝╚═════╝     ╚═════╝  ╚═════╝  ╚═════╝╚══════╝
'''


def parse_args() -> argparse.Namespace:
    """Parses command-line arguments."""

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--sphinx-build-dir',
        required=True,
        help='Directory in which to build docs',
    )
    parser.add_argument(
        '--conf', required=True, help='Path to conf.py file for Sphinx'
    )
    parser.add_argument(
        '-j',
        '--parallel',
        type=int,
        default=os.cpu_count(),
        help='Number of parallel processes to run',
    )
    parser.add_argument(
        '--gn-root', required=True, help='Root of the GN build tree'
    )
    parser.add_argument(
        '--gn-gen-root', required=True, help='Root of the GN gen tree'
    )
    parser.add_argument(
        'sources', nargs='+', help='Paths to the root level rst source files'
    )
    parser.add_argument(
        '--out-dir',
        required=True,
        help='Output directory for rendered HTML docs',
    )
    parser.add_argument(
        '--metadata',
        required=True,
        type=argparse.FileType('r'),
        help='Metadata JSON file',
    )
    parser.add_argument(
        '--google-analytics-id',
        const=None,
        help='Enables Google Analytics with the provided ID',
    )
    return parser.parse_args()


def build_docs(
    src_dir: str,
    dst_dir: str,
    parallel: int,
    google_analytics_id: str | None = None,
) -> int:
    """Runs Sphinx to render HTML documentation from a doc tree."""

    # TODO(frolv): Specify the Sphinx script from a prebuilts path instead of
    # requiring it in the tree.
    command = [
        'sphinx-build',
        '-W',
        '-j',
        str(parallel),
        '-b',
        'html',
        '-d',
        f'{dst_dir}/help',
    ]

    if google_analytics_id is not None:
        command.append(f'-Dgoogle_analytics_id={google_analytics_id}')

    command.extend([src_dir, f'{dst_dir}/html'])

    return subprocess.call(command)


def copy_doc_tree(args: argparse.Namespace) -> None:
    """Copies doc source and input files into a build tree."""

    def build_path(path):
        """Converts a source path to a filename in the build directory."""
        if path.startswith(args.gn_root):
            path = os.path.relpath(path, args.gn_root)
        elif path.startswith(args.gn_gen_root):
            path = os.path.relpath(path, args.gn_gen_root)

        return os.path.join(args.sphinx_build_dir, path)

    source_files = json.load(args.metadata)
    copy_paths = [build_path(f) for f in source_files]

    os.makedirs(args.sphinx_build_dir)
    for source_path in args.sources:
        os.link(
            source_path, f'{args.sphinx_build_dir}/{Path(source_path).name}'
        )
    os.link(args.conf, f'{args.sphinx_build_dir}/conf.py')

    # Map of directory path to list of source and destination file paths.
    dirs: dict[str, list[tuple[str, str]]] = collections.defaultdict(list)

    for source_file, copy_path in zip(source_files, copy_paths):
        dirname = os.path.dirname(copy_path)
        dirs[dirname].append((source_file, copy_path))

    for directory, file_pairs in dirs.items():
        os.makedirs(directory, exist_ok=True)
        for src, dst in file_pairs:
            os.link(src, dst)


def main() -> int:
    """Script entry point."""

    args = parse_args()

    # Clear out any existing docs for the target.
    if os.path.exists(args.sphinx_build_dir):
        shutil.rmtree(args.sphinx_build_dir)

    # TODO: b/235349854 - Printing the header causes unicode problems on
    # Windows. Disabled for now; re-enable once the root issue is fixed.
    # print(SCRIPT_HEADER)
    copy_doc_tree(args)

    # Flush all script output before running Sphinx.
    print('-' * 80, flush=True)

    return build_docs(
        args.sphinx_build_dir,
        args.out_dir,
        args.parallel,
        args.google_analytics_id,
    )


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""JavaScript and TypeScript validity check."""

from pw_presubmit import presubmit_context
from pw_presubmit.presubmit import (
    Check,
    filter_paths,
)
from pw_presubmit.presubmit_context import (
    PresubmitContext,
    PresubmitFailure,
)
from pw_presubmit.tools import log_run


@filter_paths(endswith=('.js', '.ts'))
@Check
def eslint(ctx: PresubmitContext):
    """Presubmit check that ensures JavaScript files are valid."""

    ctx.paths = presubmit_context.apply_exclusions(ctx)

    # Check if npm deps are installed.
    npm_list = log_run(['npm', 'list'])
    if npm_list.returncode != 0:
        npm_install = log_run(['npm', 'install'])
        if npm_install.returncode != 0:
            raise PresubmitFailure('npm install failed.')

    result = log_run(['npm', 'exec', 'eslint', *ctx.paths])
    if result.returncode != 0:
        raise PresubmitFailure('eslint identifed issues.')



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Console HTTP Log Server functions."""

import asyncio
import datetime
import email.utils
import logging
import mimetypes
from pathlib import Path
import socket
from threading import Thread
import webbrowser
from typing import Any, Callable

from aiohttp import web, WSMsgType

from pw_console.web_kernel import WebKernel

_LOG = logging.getLogger(__package__)


def aiohttp_server(handler: Callable) -> web.AppRunner:
    app = web.Application()
    app.add_routes([web.get('/', handler), web.get('/{name}', handler)])
    runner = web.AppRunner(app)
    return runner


def find_available_port(start_port=8080, max_retries=100) -> int:
    """Finds the next available port starting from `start_port`."""
    for port in range(start_port, start_port + max_retries):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(('localhost', port))
                return port
            except OSError:
                pass  # Port is already in use, try the next one

    raise RuntimeError(
        (
            'No available port found within the range '
            f'{start_port}-{start_port + max_retries}'
        )
    )


def pw_console_http_server(
    start_port: int,
    html_files: dict[str, str],
    kernel_params: dict[str, Any] | None = None,
) -> None:
    try:
        handler = WebHandler(html_files=html_files, kernel_params=kernel_params)
        handler.start_web_socket_streaming_responder_thread()
        runner = aiohttp_server(handler.handle_request)
        port = find_available_port(start_port)
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(runner.setup())
        site = web.TCPSite(runner, 'localhost', port)
        loop.run_until_complete(site.start())
        url = f'http://localhost:{port}'
        print(url)
        webbrowser.open(url)
        loop.run_forever()
    except KeyboardInterrupt:
        _LOG.info('Shutting down...')
        handler.stop_web_socket_streaming_responder_thread()
        loop.stop()


class WebHandler:
    """Request handler that serves files from pw_console.html package data."""

    def __init__(
        self,
        html_files: dict[str, str],
        kernel_params: dict[str, Any] | None = None,
    ) -> None:
        self.html_files = html_files
        self.date_modified = email.utils.formatdate(
            datetime.datetime.now().timestamp(), usegmt=True
        )
        self.kernel_params: dict[str, Any] = {}
        if kernel_params:
            self.kernel_params = kernel_params

        self.web_socket_streaming_responder_loop = asyncio.new_event_loop()

    def _web_socket_streaming_responder_thread_entry(self):
        """Entry point for the web socket logging handlers thread."""
        asyncio.set_event_loop(self.web_socket_streaming_responder_loop)
        self.web_socket_streaming_responder_loop.run_forever()

    def start_web_socket_streaming_responder_thread(self):
        """Start thread for handling log messages to web socket responses."""
        thread = Thread(
            target=self._web_socket_streaming_responder_thread_entry,
            args=(),
            daemon=True,
        )
        thread.start()

    def stop_web_socket_streaming_responder_thread(self):
        self.web_socket_streaming_responder_loop.call_soon_threadsafe(
            self.web_socket_streaming_responder_loop.stop
        )

    async def handle_request(
        self, request: web.Request
    ) -> web.Response | web.WebSocketResponse:
        _LOG.debug(
            '%s: %s',
            request.remote,
            request.raw_path,
        )

        path = request.path
        if path == '/':
            path = '/console.html'

        if path == '/ws':
            return await self.handle_websocket(request)

        if path not in self.html_files:
            return web.Response(status=404, text='File not found')

        content: bytes = self.html_files[path].encode('utf-8')
        content_type = 'application/octet-stream'
        mime_guess, _ = mimetypes.guess_type(Path(path).name)
        if mime_guess:
            content_type = mime_guess

        return web.Response(
            body=content,
            content_type=content_type,
            headers={
                'Content-Length': str(len(content)),
                'Last-Modified': self.date_modified,
            },
        )

    async def handle_websocket(
        self, request: web.Request
    ) -> web.WebSocketResponse:
        """Handle a websocket connection request by creating a new kernel."""
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        kernel = WebKernel(
            ws, self.kernel_params, self.web_socket_streaming_responder_loop
        )
        try:
            async for msg in ws:
                if msg.type == WSMsgType.TEXT:
                    if msg.data == 'close':
                        return ws
                    response = await kernel.handle_request(msg.data)
                    await ws.send_str(response)
                elif msg.type == WSMsgType.ERROR:
                    _LOG.warning(
                        'ws connection closed with exception: %s',
                        ws.exception(),
                    )
        finally:
            kernel.handle_disconnect()
            await ws.close()

        return ws



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Custom role for creating bug links."""

from docutils import nodes
from docutils.parsers.rst import roles

from sphinx.application import Sphinx


def bug_role(
    role,  # pylint: disable=unused-argument
    rawtext,
    text,
    lineno,
    inliner,
    options=None,
    content=None,  # pylint: disable=unused-argument
):
    options = roles.normalized_role_options(options)
    try:
        bug = int(nodes.unescape(text))
        if bug < 1:
            raise ValueError
    except ValueError:
        msg = inliner.reporter.error(
            'Bug number must be a number greater than or equal to 1; '
            '"%s" is invalid.' % text,
            line=lineno,
        )
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]

    ref = 'https://pwbug.dev/{:d}'.format(bug)
    node = nodes.reference(rawtext, 'b/{:d}'.format(bug), refuri=ref, **options)
    return [node], []


def setup(app: Sphinx) -> dict[str, bool]:
    app.add_role('bug', bug_role)
    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Formatting library core."""

import abc
from dataclasses import dataclass
import difflib
import logging
from pathlib import Path
from typing import Callable, Iterable, Iterator

from pw_cli.tool_runner import ToolRunner, BasicSubprocessRunner


_LOG: logging.Logger = logging.getLogger(__name__)


def _ensure_newline(orig: str) -> str:
    """Adds a warning and newline to any file without a trailing newline."""

    if orig.endswith('\n'):
        return orig
    return orig + '\nNo newline at end of file\n'


def simple_diff(path: Path, original: str, formatted: str) -> str:
    """Produces a diff of the contents of two files."""

    original = _ensure_newline(original)
    formatted = _ensure_newline(formatted)
    return ''.join(
        difflib.unified_diff(
            original.splitlines(keepends=True),
            formatted.splitlines(keepends=True),
            f'{path}  (original)',
            f'{path}  (reformatted)',
        )
    )


DiffCallback = Callable[[Path, str, str], str]
"""The callback type for producing diffs.

Arugments:
    path: File path of the file being diffed.
    original: The contents of the original file, as a string.
    formatted: The contents of the formatted file, as a string.

Returns:
    A human readable diff as a string.
"""


@dataclass(frozen=True)
class FormattedFileContents:
    """The result of running a code formatter on the contents of a file.

    This type is returned by in-memory formatting check operations.

    Attributes:
        ok: A boolean indicating whether or not formatting was successful.
        formatted_file_contents: The contents of the resulting formatted file
            as bytes.
        error_message: A string containing any errors or warnings produced by
            the formatting process.
    """

    ok: bool
    formatted_file_contents: bytes
    error_message: str | None


@dataclass(frozen=True)
class FormattedDiff:
    """The resulting diff of applying a code formatter to a file.

    Attributes:
        ok: A boolean indicating whether or not formatting was successful.
        diff: The resulting diff of applying code formatting, as a
            human-readable string.
        error_message: A string containing any errors or warnings produced by
            the formatting process.
        file_path: The path of the corresponding file that produced this diff.
    """

    ok: bool
    diff: str
    error_message: str | None
    file_path: Path


@dataclass(frozen=True)
class FormatFixStatus:
    """The status of running a code formatter in-place on a file.

    This type is returned by in-place formatting fix operations.

    Attributes:
        ok: A boolean indicating whether or not formatting was successful.
        error_message: A string containing any errors or warnings produced by
            the formatting process.
    """

    ok: bool
    error_message: str | None


class FileChecker(abc.ABC):
    """Abstract class for a code format check tool.

    This class does not have the ability to apply formatting to files, and
    instead only allows in-memory checks to produce expected resulting diffs.

    Attributes:
        run_tool: The :py:class:`pw_presubmit.format.core.ToolRunner` to use
            when calling out to subprocesses.
        diff_tool: The :py:attr:`pw_presubmit.format.core.DiffCallback` to use
            when producing formatting diffs.
    """

    def __init__(
        self,
        tool_runner: ToolRunner = BasicSubprocessRunner(),
        diff_tool: DiffCallback = simple_diff,
    ):
        # Always call `self.run_tool` rather than `subprocess.run`, as it allows
        # injection of tools and other environment-specific handlers.
        self.run_tool = tool_runner
        self.diff_tool = diff_tool

    @abc.abstractmethod
    def format_file_in_memory(
        self, file_path: Path, file_contents: bytes
    ) -> FormattedFileContents:
        """Returns the formatted version of a file as in-memory bytes.

        ``file_path`` and ``file_content`` represent the same file. Both are
        provided for convenience. Use ``file_path`` if you can do so without
        modifying the file, or use ``file_contents`` if the formatting tool
        provides a mechanism for formatting the file by piping it to stdin.

        Any subprocess calls should be initiated with ``self.run_tool()`` to
        enable testing and injection of tools and tool wrappers.

        **WARNING**: A :py:class:`pw_presubmit.format.core.FileChecker` must
        **never** modify the file at``file_path``.

        Returns:
            A populated
            :py:class:`pw_presubmit.format.core.FormattedFileContents` that
            contains either the result of formatting the file, or an error
            message.
        """

    def get_formatting_diff(
        self, file_path: Path, dry_run: bool = False
    ) -> FormattedDiff | None:
        """Returns a diff comparing a file to its formatted version.

        If ``dry_run`` is ``True``, the diff will always be ``None``.

        Returns:
            ``None`` if there is no difference after formatting **OR** if
            ``dry_run`` is enabled. Otherwise, a
            :py:class:`pw_presubmit.format.core.FormattedDiff` is returned
            containing either a diff or an error.
        """
        original = file_path.read_bytes()

        formatted = self.format_file_in_memory(file_path, original)

        if not formatted.ok:
            return FormattedDiff(
                diff='',  # Don't try to diff.
                ok=False,
                file_path=file_path,
                error_message=formatted.error_message,
            )

        if dry_run:
            return None

        # No difference found.
        if formatted.formatted_file_contents == original:
            return None

        return FormattedDiff(
            diff=self.diff_tool(
                file_path,
                original.decode(errors='replace'),
                formatted.formatted_file_contents.decode(errors='replace'),
            ),
            file_path=file_path,
            error_message=formatted.error_message,
            ok=True,
        )

    def get_formatting_diffs(
        self, paths: Iterable[Path], dry_run: bool = False
    ) -> Iterator[FormattedDiff]:
        """Checks the formatting of many files without modifying them.

        This method may be overridden to optimize for formatters that allow
        checking multiple files in a single invocation, though you may need
        to do additional parsing to produce diffs or error messages associated
        with each file path.

        Returns:
            An iterator of :py:class:`pw_presubmit.format.core.FormattingDiff`
            objects for each file with identified formatting issues.
        """

        for path in paths:
            diff = self.get_formatting_diff(path, dry_run)
            if diff is not None:
                yield diff


class FileFormatter(FileChecker):
    """Abstract class for a code format fix tool."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abc.abstractmethod
    def format_file(self, file_path: Path) -> FormatFixStatus:
        """Formats the provided file in-place.

        Any subprocess calls should be initiated with ``self.run_tool()`` to
        enable testing and injection of tools and tool wrappers.

        Returns:
            A FormatFixStatus that contains relevant errors/warnings.
        """

    def format_files(
        self, paths: Iterable[Path], keep_warnings: bool = True
    ) -> Iterator[tuple[Path, FormatFixStatus]]:
        """Formats the provided files and fixes them in-place.

        All files must be updated to contain the formatted version. If errors
        are encountered along the way, they should be collected and returned as
        a dictionary that maps the path of the file to a string that
        describes the errors encountered while processing that file.

        Any subprocess calls should be initiated with ``self.run_tool()`` to
        enable testing and injection of tools and tool wrappers.

        This method may be overridden to optimize for formatters that allow
        formatting multiple files in a single invocation, though you may need
        to do additional parsing to associate error messages with the paths of
        the files that produced them.

        Returns:
            An iterator of ``Path`` and
            :py:class:`pw_presubmit.format.core.FormatFixStatus` pairs for each
            file that was not successfully formatted. If ``keep_warnings`` is
            ``True``, any successful format operations with warnings will also
            be returned.
        """

        for file_path in paths:
            status = self.format_file(file_path)
            if not status.ok or (status.error_message and keep_warnings):
                yield (file_path, status)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_unit_test_proto/unit_test.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"pw_unit_test_proto/unit_test.proto\x12\x0cpw.unit_test\"N\n\x12TestCaseDescriptor\x12\x12\n\nsuite_name\x18\x01 \x01(\t\x12\x11\n\ttest_name\x18\x02 \x01(\t\x12\x11\n\tfile_name\x18\x03 \x01(\t\"m\n\x13TestCaseExpectation\x12\x12\n\nexpression\x18\x01 \x01(\t\x12\x1c\n\x14\x65valuated_expression\x18\x02 \x01(\t\x12\x13\n\x0bline_number\x18\x03 \x01(\r\x12\x0f\n\x07success\x18\x04 \x01(\x08\"\x0e\n\x0cTestRunStart\"O\n\nTestRunEnd\x12\x0e\n\x06passed\x18\x01 \x01(\r\x12\x0e\n\x06\x66\x61iled\x18\x02 \x01(\r\x12\x0f\n\x07skipped\x18\x03 \x01(\r\x12\x10\n\x08\x64isabled\x18\x04 \x01(\r\"\xef\x02\n\x05\x45vent\x12\x34\n\x0etest_run_start\x18\x01 \x01(\x0b\x32\x1a.pw.unit_test.TestRunStartH\x00\x12\x30\n\x0ctest_run_end\x18\x02 \x01(\x0b\x32\x18.pw.unit_test.TestRunEndH\x00\x12;\n\x0ftest_case_start\x18\x03 \x01(\x0b\x32 .pw.unit_test.TestCaseDescriptorH\x00\x12\x35\n\rtest_case_end\x18\x04 \x01(\x0e\x32\x1c.pw.unit_test.TestCaseResultH\x00\x12>\n\x12test_case_disabled\x18\x05 \x01(\x0b\x32 .pw.unit_test.TestCaseDescriptorH\x00\x12\x42\n\x15test_case_expectation\x18\x06 \x01(\x0b\x32!.pw.unit_test.TestCaseExpectationH\x00\x42\x06\n\x04type\"H\n\x0eTestRunRequest\x12\"\n\x1areport_passed_expectations\x18\x01 \x01(\x08\x12\x12\n\ntest_suite\x18\x02 \x03(\t*7\n\x0eTestCaseResult\x12\x0b\n\x07SUCCESS\x10\x00\x12\x0b\n\x07\x46\x41ILURE\x10\x01\x12\x0b\n\x07SKIPPED\x10\x02\x32H\n\x08UnitTest\x12<\n\x03Run\x12\x1c.pw.unit_test.TestRunRequest\x1a\x13.pw.unit_test.Event\"\x00\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_unit_test_proto.unit_test_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_TESTCASERESULT']._serialized_start=784
  _globals['_TESTCASERESULT']._serialized_end=839
  _globals['_TESTCASEDESCRIPTOR']._serialized_start=52
  _globals['_TESTCASEDESCRIPTOR']._serialized_end=130
  _globals['_TESTCASEEXPECTATION']._serialized_start=132
  _globals['_TESTCASEEXPECTATION']._serialized_end=241
  _globals['_TESTRUNSTART']._serialized_start=243
  _globals['_TESTRUNSTART']._serialized_end=257
  _globals['_TESTRUNEND']._serialized_start=259
  _globals['_TESTRUNEND']._serialized_end=338
  _globals['_EVENT']._serialized_start=341
  _globals['_EVENT']._serialized_end=708
  _globals['_TESTRUNREQUEST']._serialized_start=710
  _globals['_TESTRUNREQUEST']._serialized_end=782
  _globals['_UNITTEST']._serialized_start=841
  _globals['_UNITTEST']._serialized_end=913
# @@protoc_insertion_point(module_scope)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Parses Bazel rules from a local Bazel workspace."""

import json
import re
import subprocess

from pathlib import PurePath, PurePosixPath
from typing import (
    Any,
    Callable,
    Iterable,
)

BazelValue = bool | int | str | list[str] | dict[str, str]

LABEL_PAT = re.compile(r'^(?:(?:@([a-zA-Z0-9_]*))?//([^:]*))?(?::([^:]+))?$')


class ParseError(Exception):
    """Raised when a Bazel query returns data that can't be parsed."""


class BazelLabel:
    """Represents a Bazel target identifier."""

    def __init__(
        self,
        label_str: str,
        repo: str | None = None,
        package: str | None = None,
    ) -> None:
        """Creates a Bazel label.

        This method will attempt to parse the repo, package, and target portion
        of the given label string. If the repo portion of the label is omitted,
        it will use the given repo, if provided. If the repo and the package
        portions of the label are omitted, it will use the given package, if
        provided. If the target portion is omitted, the last segment of the
        package will be used.

        Args:
            label_str: Bazel label string, like "@repo//pkg:target".
            repo: Repo to use if omitted from label, e.g. as in "//pkg:target".
            package: Package to use if omitted from label, e.g. as in ":target".
        """
        match = re.match(LABEL_PAT, label_str)
        if not label_str or not match:
            raise ParseError(f'invalid label: "{label_str}"')
        if match.group(1) or not package:
            package = ''
        if match.group(1):
            self._repo = match.group(1)
        else:
            assert repo
            self._repo = repo
        self._package = match.group(2) if match.group(2) else package
        self._target = (
            match.group(3)
            if match.group(3)
            else PurePosixPath(self._package).name
        )

    def __str__(self) -> str:
        """Canonical representation of a Bazel label."""
        return f'@{self._repo}//{self._package}:{self._target}'

    def repo(self) -> str:
        """Returns the repository identifier associated with this label."""
        return self._repo

    def package(self) -> str:
        """Returns the package path associated with this label."""
        return self._package

    def target(self) -> str:
        """Returns the target name associated with this label."""
        return self._target


def parse_invalid(attr: dict[str, Any]) -> BazelValue:
    """Raises an error that a type is unrecognized."""
    attr_type = attr['type']
    raise ParseError(f'unknown type: {attr_type}, expected one of {BazelValue}')


class BazelRule:
    """Represents a Bazel rule as parsed from the query results."""

    def __init__(self, kind: str, label: BazelLabel) -> None:
        """Create a Bazel rule.

        Args:
            kind: The type of Bazel rule, e.g. cc_library.
            label: A Bazel label corresponding to this rule.
        """
        self._kind = kind
        self._label = label
        self._attrs: dict[str, BazelValue] = {}
        self._types: dict[str, str] = {}

    def kind(self) -> str:
        """Returns this rule's target type."""
        return self._kind

    def label(self) -> BazelLabel:
        """Returns this rule's Bazel label."""
        return self._label

    def parse_attrs(self, attrs: Iterable[dict[str, Any]]) -> None:
        """Maps JSON data from a bazel query into this object.

        Args:
            attrs: A dictionary of attribute names and values for the Bazel
                rule. These should match the output of
                `bazel cquery ... --output=jsonproto`.
        """
        attr_parsers: dict[str, Callable[[dict[str, Any]], BazelValue]] = {
            'boolean': lambda attr: attr.get('booleanValue', False),
            'integer': lambda attr: int(attr.get('intValue', '0')),
            'string': lambda attr: attr.get('stringValue', ''),
            'label': lambda attr: attr.get('stringValue', ''),
            'string_list': lambda attr: attr.get('stringListValue', []),
            'label_list': lambda attr: attr.get('stringListValue', []),
            'string_dict': lambda attr: {
                p['key']: p['value'] for p in attr.get('stringDictValue', [])
            },
        }
        for attr in attrs:
            if 'explicitlySpecified' not in attr:
                continue
            if not attr['explicitlySpecified']:
                continue
            try:
                attr_name = attr['name']
            except KeyError:
                raise ParseError(
                    f'missing "name" in {json.dumps(attr, indent=2)}'
                )
            try:
                attr_type = attr['type'].lower()
            except KeyError:
                raise ParseError(
                    f'missing "type" in {json.dumps(attr, indent=2)}'
                )

            attr_parser = attr_parsers.get(attr_type, parse_invalid)
            self._attrs[attr_name] = attr_parser(attr)
            self._types[attr_name] = attr_type

    def has_attr(self, attr_name: str) -> bool:
        """Returns whether the rule has an attribute of the given name.

        Args:
            attr_name: The name of the attribute.
        """
        return attr_name in self._attrs

    def attr_type(self, attr_name: str) -> str:
        """Returns the type of an attribute according to Bazel.

        Args:
            attr_name: The name of the attribute.
        """
        return self._types[attr_name]

    def get_bool(self, attr_name: str) -> bool:
        """Gets the value of a boolean attribute.

        Args:
            attr_name: The name of the boolean attribute.
        """
        val = self._attrs.get(attr_name, False)
        assert isinstance(val, bool)
        return val

    def get_int(self, attr_name: str) -> int:
        """Gets the value of an integer attribute.

        Args:
            attr_name: The name of the integer attribute.
        """
        val = self._attrs.get(attr_name, 0)
        assert isinstance(val, int)
        return val

    def get_str(self, attr_name: str) -> str:
        """Gets the value of a string attribute.

        Args:
            attr_name: The name of the string attribute.
        """
        val = self._attrs.get(attr_name, '')
        assert isinstance(val, str)
        return val

    def get_list(self, attr_name: str) -> list[str]:
        """Gets the value of a string list attribute.

        Args:
            attr_name: The name of the string list attribute.
        """
        val = self._attrs.get(attr_name, [])
        assert isinstance(val, list)
        return val

    def get_dict(self, attr_name: str) -> dict[str, str]:
        """Gets the value of a string list attribute.

        Args:
            attr_name: The name of the string list attribute.
        """
        val = self._attrs.get(attr_name, {})
        assert isinstance(val, dict)
        return val

    def set_attr(self, attr_name: str, value: BazelValue) -> None:
        """Sets the value of an attribute.

        Args:
            attr_name: The name of the attribute.
            value: The value to set.
        """
        self._attrs[attr_name] = value

    def filter_attr(self, attr_name: str, remove: Iterable[str]) -> None:
        """Removes values from a list attribute.

        Args:
            attr_name: The name of the attribute.
            remove: The values to remove.
        """
        values = self.get_list(attr_name)
        self._attrs[attr_name] = [v for v in values if v not in remove]

    def generate(self) -> Iterable[str]:
        """Yields a sequence of strings describing the rule in Bazel."""
        yield f'{self._kind}('
        yield f'    name = "{self._label.target()}",'
        for name in sorted(self._attrs.keys()):
            if name == 'name':
                continue
            attr_type = self._types[name]
            if attr_type == 'boolean':
                yield f'    {name} = {self.get_bool(name)},'
            elif attr_type == 'integer':
                yield f'    {name} = {self.get_int(name)},'
            elif attr_type == 'string':
                yield f'    {name} = "{self.get_str(name)}",'
            elif attr_type == 'string_list' or attr_type == 'label_list':
                strs = self.get_list(name)
                if len(strs) == 1:
                    yield f'    {name} = ["{strs[0]}"],'
                elif len(strs) > 1:
                    yield f'    {name} = ['
                    for s in strs:
                        yield f'        "{s}",'
                    yield '    ],'
            elif attr_type == 'string_dict':
                str_dict = self.get_dict(name)
                yield f'    {name} = {{'
                for k, v in str_dict.items():
                    yield f'        {k} = "{v}",'
                yield '    },'
        yield ')'


class BazelWorkspace:
    """Represents a local instance of a Bazel repository.

    Attributes:
        defaults: Attributes automatically applied to every rule in the
                  workspace, which may be ignored when parsing.
        generate: Indicates whether GN should be automatically generated for
                  this workspace or not.
        targets:  A list of the Bazel targets to parse, along with their
                  dependencies.
    """

    def __init__(
        self, repo: str, source_dir: PurePath | None, fetch: bool = True
    ) -> None:
        """Creates an object representing a Bazel workspace at the given path.

        Args:
            repo: The Bazel repository name, like "com_google_pigweed".
            source_dir: Path to the local instance of a Bazel workspace.
        """
        self.defaults: dict[str, list[str]] = {}
        self.generate = True
        self.targets: list[str] = []
        self.options: dict[str, Any] = {}
        self._fetched = False
        self._repo: str = repo
        self._revisions: dict[str, str] = {}
        self._rules: dict[str, BazelRule] = {}
        self._source_dir = source_dir

        # Make sure the workspace has up-to-date objects and refs.
        if fetch:
            self._git('fetch')

    def repo(self) -> str:
        """Returns the Bazel repository name for this workspace."""
        return self._repo

    def get_http_archives(self) -> Iterable[BazelRule]:
        """Returns the http_archive rules from a workspace's WORKSPACE file."""
        if not self.generate:
            return
        for result in self._query('kind(http_archive, //external:*)'):
            if result['type'] != 'RULE':
                continue
            yield self._make_rule(result['rule'])

    def get_rules(self, labels: list[BazelLabel]) -> Iterable[BazelRule]:
        """Returns a rule matching the given label."""
        if not self.generate:
            return
        needed: list[str] = []
        for label in labels:
            print(f'Examining [{label}]                             ')
            short = f'//{label.package()}:{label.target()}'
            rule = self._rules.get(short, None)
            if rule:
                yield rule
                continue
            needed.append(short)
        flags = [f'--{label}={value}' for label, value in self.options.items()]
        results = list(self._cquery('+'.join(needed), flags))
        for result in results:
            rule_data = result['target']['rule']
            yield self._make_rule(rule_data)

    def revision(self, commitish: str = 'HEAD') -> str:
        """Returns the revision digest of the workspace's git commit-ish."""
        try:
            return self._git('rev-parse', commitish)
        except ParseError:
            pass
        tags = self._git('tag', '--list').split()
        tag = min([tag for tag in tags if commitish in tag], key=len)
        return self._git('rev-parse', tag)

    def timestamp(self, revision: str) -> str:
        """Returns the timestamp of the workspace's git commit-ish."""
        return self._git('show', '--no-patch', '--format=%ci', revision)

    def url(self) -> str:
        """Returns the git URL of the workspace."""
        return self._git('remote', 'get-url', 'origin')

    def _cquery(self, expr: str, flags: list[str]) -> Iterable[Any]:
        """Invokes `bazel cquery` with the given selector."""
        result = self._bazel('cquery', expr, *flags, '--output=jsonproto')
        return json.loads(result)['results']

    def _query(self, expr: str) -> Iterable[Any]:
        """Invokes `bazel query` with the given selector."""
        results = self._bazel('query', expr, '--output=streamed_jsonproto')
        return [json.loads(result) for result in results.split('\n')]

    def _make_rule(self, rule_data: Any) -> BazelRule:
        """Make a BazelRule from JSON data returned by query or cquery."""
        short = rule_data['name']
        label = BazelLabel(short, repo=self._repo)
        rule = BazelRule(rule_data['ruleClass'], label)
        rule.parse_attrs(rule_data['attribute'])
        for attr_name, values in self.defaults.items():
            rule.filter_attr(attr_name, values)
        self._rules[short] = rule
        return rule

    def _bazel(self, *args: str) -> str:
        """Execute a Bazel command in the workspace."""
        return self._exec('bazel', *args, '--noshow_progress', check=False)

    def _git(self, *args: str) -> str:
        """Execute a git command in the workspace."""
        return self._exec('git', *args, check=True)

    def _exec(self, *args: str, check=True) -> str:
        """Execute a command in the workspace."""
        try:
            result = subprocess.run(
                list(args),
                cwd=self._source_dir,
                check=check,
                capture_output=True,
            )
            if result.stdout:
                # The extra 'str()' facilitates MagicMock.
                return str(result.stdout.decode('utf-8')).strip()
            if check:
                return ''
            errmsg = result.stderr.decode('utf-8')
        except subprocess.CalledProcessError as error:
            errmsg = error.stderr.decode('utf-8')
        cmdline = ' '.join(list(args))
        raise ParseError(
            f'{self._repo} failed to exec '
            + f'`cd {self._source_dir} && {cmdline}`: {errmsg}'
        )



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Generates additional documentation for SEEDs."""

import argparse
import json
from pathlib import Path


def _parse_args() -> argparse.Namespace:
    """Parses command-line arguments."""

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--output',
        type=Path,
        required=True,
        help='Output file to write',
    )
    parser.add_argument(
        'seed_metadata',
        nargs='+',
        help='JSON file containing information about a SEED',
    )

    return parser.parse_args()


_STATUS_BADGES = {
    'draft': 'bdg-primary-line',
    'open for comments': 'bdg-primary',
    'last call': 'bdg-warning',
    'accepted': 'bdg-success',
    'rejected': 'bdg-danger',
    'deprecated': 'bdg-secondary',
    'superseded': 'bdg-info',
    'on hold': 'bdg-secondary-line',
    'meta': 'bdg-success-line',
}


def _status_badge(status: str) -> str:
    badge = _STATUS_BADGES.get(status.lower().strip(), 'bdg-primary')
    return f':{badge}:`{status}`'


def _main():
    args = _parse_args()

    # Table containing entries for each SEED.
    seed_table = [
        '\n.. list-table::',
        '   :widths: auto',
        '   :header-rows: 1\n',
        '   * - Number',
        '     - Title',
        '     - Status',
        '     - Author',
        '     - Facilitator',
    ]

    # RST toctree including each SEED's source file.
    seed_toctree = []

    for metadata_file in args.seed_metadata:
        with open(metadata_file, 'r') as file:
            meta = json.load(file)

            if 'changelist' in meta:
                # The SEED has not yet been merged and points to an active CL.
                change_link = (
                    'https://pigweed-review.googlesource.com'
                    f'/c/pigweed/pigweed/+/{meta["changelist"]}'
                )
                title = f'`{meta["title"]} <{change_link}>`__'
                seed_toctree.append(
                    f'{meta["number"]}: {meta["title"]}<{change_link}>',
                )
            else:
                # The SEED document is in the source tree.
                title = f':ref:`{meta["title"]} <seed-{meta["number"]}>`'
                seed_toctree.append(Path(meta["rst_file"]).stem)

            facilitator = meta.get('facilitator', 'Unassigned')

            seed_table.extend(
                [
                    f'   * - {meta["number"]}',
                    f'     - {title}',
                    f'     - {_status_badge(meta["status"])}',
                    f'     - {meta["author"]}',
                    f'     - {facilitator}',
                ]
            )

    table = '\n'.join(seed_table)

    # Hide the toctree it is only used for the sidebar.
    toctree_lines = '\n  '.join(seed_toctree)

    # fmt: off
    toctree = (
        '.. toctree::\n'
        '  :hidden:\n'
        '\n  '
        f'{toctree_lines}'
    )
    # fmt: on

    args.output.write_text(f'{table}\n\n{toctree}')


if __name__ == '__main__':
    _main()



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Docs widget that provides up-to-date info about the next Pigweed Live."""


import datetime
import sys

from docutils import nodes
from docutils.parsers.rst import Directive
from sphinx.application import Sphinx

try:
    import pytz  # type: ignore

    PYTZ_AVAILABLE = True
except ImportError:
    PYTZ_AVAILABLE = False


class PigweedLiveDirective(Directive):
    """Generates the up-to-date Pigweed Live info."""

    datetime_format = '%Y-%m-%d %H:%M:%S'
    # TODO: b/303859828 - Update this data sometime between 2024-09-23
    # and 2024-10-07.
    meetings = [
        '2024-07-01 13:00:00',
        '2024-07-29 13:00:00',
        '2024-08-26 13:00:00',
        '2024-09-23 13:00:00',
        '2024-10-21 13:00:00',
        '2024-11-18 13:00:00',
        '2024-12-16 13:00:00',
    ]
    timezone = pytz.timezone('US/Pacific')

    def run(self) -> list[nodes.Node]:
        return [self._make_paragraph()]

    def _make_paragraph(self) -> nodes.Node:
        next_meeting = self._find_next_meeting()
        paragraph = nodes.paragraph()
        paragraph += nodes.Text('Our next Pigweed Live is ')
        meeting_text = nodes.strong()
        meeting_text += nodes.Text(next_meeting)
        paragraph += meeting_text
        paragraph += nodes.Text(
            (
                ". Please join us to discuss what's new in Pigweed and "
                "anything else Pigweed-related that's on your mind. "
            )
        )
        refuri = 'https://groups.google.com/g/pigweed'
        link = nodes.reference(refuri=refuri)
        link += nodes.Text('Join our mailing list')
        paragraph += link
        paragraph += nodes.Text(" to receive an invite to the next meeting.")
        return paragraph

    def _find_next_meeting(self) -> str:
        current_datetime = self.timezone.localize(datetime.datetime.now())
        next_meeting = None
        for meeting in self.meetings:
            unlocalized_datetime = datetime.datetime.strptime(
                meeting, self.datetime_format
            )
            meeting_datetime = self.timezone.localize(unlocalized_datetime)
            if current_datetime > meeting_datetime:
                continue
            next_meeting = meeting_datetime
            break
        if next_meeting is None:
            sys.exit(
                'ERROR: Pigweed Live meeting data needs to be updated. '
                'Update the `meetings` list in `PigweedLiveDirective`. '
                'See b/303859828.'
            )
        else:
            date = next_meeting.strftime('%a %b %d, %Y')
            hour = next_meeting.strftime('%I%p').lstrip('0')
            timezone = 'PDT' if next_meeting.dst() else 'PST'
            return f'{date} {hour} ({timezone})'


def setup(app: Sphinx) -> dict[str, bool]:
    """Initialize the directive."""
    if PYTZ_AVAILABLE:
        app.add_directive('pigweed-live', PigweedLiveDirective)
    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Tools to analyze RISCV CPU state context captured during an exception."""

from typing import Optional

from pw_cpu_exception_risc_v_protos import cpu_state_pb2
import pw_symbolizer

# These registers are symbolized when dumped.
_SYMBOLIZED_REGISTERS = (
    'mepc',
    'ra',
)


class RiscvExceptionAnalyzer:
    """This class provides helper functions to dump a RiscvCpuState proto."""

    def __init__(
        self, cpu_state, symbolizer: Optional[pw_symbolizer.Symbolizer] = None
    ):
        self._cpu_state = cpu_state
        self._symbolizer = symbolizer

    def dump_registers(self) -> str:
        """Dumps all captured CPU registers as a multi-line string."""
        registers = []
        for field in self._cpu_state.DESCRIPTOR.fields:
            if self._cpu_state.HasField(field.name):
                register_value = getattr(self._cpu_state, field.name)
                register_str = f'{field.name:<10} 0x{register_value:08x}'
                if (
                    self._symbolizer is not None
                    and field.name in _SYMBOLIZED_REGISTERS
                ):
                    symbol = self._symbolizer.symbolize(register_value)
                    if symbol.name:
                        register_str += f' {symbol}'
                registers.append(register_str)
        return '\n'.join(registers)

    def __str__(self):
        dump = []
        dump.extend(
            (
                'All registers:',
                self.dump_registers(),
            )
        )
        return '\n'.join(dump)


def process_snapshot(
    serialized_snapshot: bytes,
    symbolizer: Optional[pw_symbolizer.Symbolizer] = None,
) -> str:
    """Returns the stringified result of a SnapshotCpuStateOverlay message run
    though a RiscvExceptionAnalyzer.
    """
    snapshot = cpu_state_pb2.SnapshotCpuStateOverlay()
    snapshot.ParseFromString(serialized_snapshot)

    if snapshot.HasField('riscv_cpu_state'):
        state_analyzer = RiscvExceptionAnalyzer(
            snapshot.riscv_cpu_state, symbolizer
        )
        return f'{state_analyzer}\n'

    return ''



#!/usr/bin/env python3
# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""
Generates json trace files viewable using chrome://tracing using RPCs from a
connected trace service.

Example usage:
python pw_console/py/pw_console/trace_client.py
  -o trace.json
  -t out/host_device_simulator.speed_optimized/obj/pw_system/bin/system_example
"""

import argparse
import logging
import sys
from pathlib import Path
from types import ModuleType


from pw_transfer import transfer_pb2
from pw_log.proto import log_pb2
from pw_trace_protos import trace_service_pb2
from pw_trace import trace
from pw_trace_tokenized import trace_tokenized
import pw_transfer
from pw_file import file_pb2
from pw_hdlc import rpc
from pw_stream import stream_readers
from pw_system import device_tracing
from pw_tokenizer import detokenize
from pw_console import socket_client


_LOG = logging.getLogger(__package__)
_LOG.level = logging.DEBUG
_LOG.addHandler(logging.StreamHandler(sys.stdout))


def start_tracing_on_device(client):
    """Start tracing on the device"""
    service = client.rpcs.pw.trace.proto.TraceService
    service.Start()


def stop_tracing_on_device(client):
    """Stop tracing on the device"""
    service = client.rpcs.pw.trace.proto.TraceService
    return service.Stop()


def list_files_on_device(client):
    """List files on the device"""
    service = client.rpcs.pw.file.FileSystem
    return service.List()


def delete_file_on_device(client, path):
    """Delete a file on the device"""
    service = client.rpcs.pw.file.FileSystem
    req = file_pb2.DeleteRequest(path=path)
    return service.Delete(req)


def _parse_args():
    """Parse and return command line arguments."""

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument('-d', '--device', help='the serial port to use')
    parser.add_argument(
        '-b',
        '--baudrate',
        type=int,
        default=115200,
        help='The baud rate to use',
    )
    group.add_argument(
        '-s',
        '--socket-addr',
        type=str,
        default='default',
        help=(
            'Use socket to connect to server, type default for '
            'localhost:33000, or manually input the server address:port'
        ),
    )
    parser.add_argument(
        '-o',
        '--trace_output',
        dest='trace_output_file',
        help='The json file to which to write the output.',
    )
    parser.add_argument(
        '-t',
        '--trace_token_database',
        help='Databases (ELF, binary, or CSV) to use to lookup trace tokens.',
    )
    parser.add_argument(
        '-f',
        '--ticks_per_second',
        type=int,
        dest='ticks_per_second',
        help='The clock rate of the trace events.',
    )
    parser.add_argument(
        '--time_offset',
        type=int,
        dest='time_offset',
        default=0,
        help='Time offset (us) of the trace events (Default 0).',
    )
    parser.add_argument(
        '--channel-id',
        type=int,
        dest='channel_id',
        default=rpc.DEFAULT_CHANNEL_ID,
        help='Channel ID used in RPC communications.',
    )
    return parser.parse_args()


def _main(args) -> int:
    detokenizer = detokenize.AutoUpdatingDetokenizer(
        args.trace_token_database + '#trace'
    )
    detokenizer.show_errors = True

    socket_impl = socket_client.SocketClient
    try:
        socket_device = socket_impl(args.socket_addr)
        reader = stream_readers.SelectableReader(socket_device)
        write = socket_device.write
    except ValueError:
        _LOG.exception('Failed to initialize socket at %s', args.socket_addr)
        return 1

    protos: list[ModuleType | Path] = [
        log_pb2,
        file_pb2,
        transfer_pb2,
        trace_service_pb2,
    ]

    with reader:
        device_client = device_tracing.DeviceWithTracing(
            args.channel_id,
            reader,
            write,
            protos,
            detokenizer=detokenizer,
            timestamp_decoder=None,
            rpc_timeout_s=5,
            use_rpc_logging=True,
            use_hdlc_encoding=True,
            ticks_per_second=args.ticks_per_second,
        )

        with device_client:
            _LOG.info('Starting tracing')
            start_tracing_on_device(device_client)

            _LOG.info('Stopping tracing')
            file_id = stop_tracing_on_device(device_client)
            _LOG.info('Trace file id = %d', file_id.response.file_id)

            _LOG.info('Listing Files')
            stream_response = list_files_on_device(device_client)

            if not stream_response.status.ok():
                _LOG.error('Failed to list files %s', stream_response.status)
                return 1

            for list_response in stream_response.responses:
                for file in list_response.paths:
                    _LOG.info('Transfering File: %s', file.path)
                    try:
                        data = device_client.transfer_manager.read(file.file_id)
                        events = trace_tokenized.get_trace_events(
                            [detokenizer.database],
                            data,
                            device_client.ticks_per_second,
                            args.time_offset,
                        )
                        json_lines = trace.generate_trace_json(events)
                        trace_tokenized.save_trace_file(
                            json_lines, args.trace_output_file
                        )
                    except pw_transfer.Error as err:
                        print('Failed to read:', err.status)

                    _LOG.info('Deleting File: %s', file.path)
                    delete_file_on_device(device_client, file.path)

            _LOG.info('All trace transfers completed successfully')

    return 0


if __name__ == '__main__':
    _main(_parse_args())



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""snapshot handler"""

import dataclasses
import logging
from io import StringIO

from pw_log import log_decoder
from pw_snapshot import processor
from pw_snapshot_protos import snapshot_pb2
from pw_tokenizer import detokenize


def _parse_snapshot(serialized_snapshot: bytes) -> snapshot_pb2.Snapshot:
    """Parse the serialized snapshot as a Snapshot proto."""
    return snapshot_pb2.Snapshot.FromString(serialized_snapshot)


def _process_logs(
    detokenizer: detokenize.Detokenizer | None, snapshot: snapshot_pb2.Snapshot
) -> str:
    """Returns the logs from the snapshot."""
    if not snapshot.logs:
        return "No captured logs\n"

    output: list[str] = [
        "Device Logs: ",
    ]

    decoded_log_stream = StringIO()
    handler = logging.StreamHandler(stream=decoded_log_stream)

    string_logger = logging.getLogger('crash_snapshot_logs')
    # only send logs to o StringIO handler
    string_logger.propagate = False
    string_logger.addHandler(handler)

    # LogStreamDecoder requires a decoded_log_handler, but it never actually
    # is invoked here, as we manually iterate the logs and call
    # parse_log_entry_proto directly.
    # pylint: disable=unused-argument
    def nop_log_handler(log: log_decoder.Log) -> None:
        return

    # pylint: enable=unused-argument

    decoder = log_decoder.LogStreamDecoder(
        decoded_log_handler=nop_log_handler,
        detokenizer=detokenizer,
        source_name='RpcDevice',
    )
    for log_entry in snapshot.logs:
        parsed_log = decoder.parse_log_entry_proto(log_entry)
        log_decoder.log_decoded_log(parsed_log, string_logger)

    output.append(decoded_log_stream.getvalue())

    return "\n".join(output)


@dataclasses.dataclass
class _CustomProcessor:
    """Snapshot processor callback handler."""

    detokenizer: detokenize.Detokenizer | None

    def __call__(self, serialized_snapshot: bytes) -> str:
        snapshot = _parse_snapshot(serialized_snapshot)

        output: list[str] = []
        output.append(_process_logs(self.detokenizer, snapshot))

        return "\n".join(output)


def decode_snapshot(
    detokenizer: detokenize.Detokenizer | None,
    serialized_snapshot: bytes,
) -> str:
    """Decodes the serialized snapshot

    Args:
      detokenizer: String detokenizer.
      serialized_snapshot: Contents of the snapshot dump file.

    Returns:
      The decoded snapshot as a string.
    """
    custom_processor = _CustomProcessor(
        detokenizer,
    )

    return processor.process_snapshots(
        serialized_snapshot=serialized_snapshot,
        detokenizer=detokenizer,
        user_processing_callback=custom_processor,
    )



#!/usr/bin/env python
# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Pigweed Zephyr argparser."""
import argparse


def add_parser_arguments(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    """Sets up an argument parser for pw zephyr."""
    parser.add_argument(
        '-v',
        '--verbose',
        dest='verbose',
        action='store_true',
        help="Enables debug logging",
    )

    subparsers = parser.add_subparsers(dest='zephyr_subcommand')
    subparsers.add_parser(
        'manifest',
        help=(
            'Prints the West manifest entries needed to add this Pigweed'
            ' instance'
        ),
    )

    return parser



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Print git-style diffs."""


import difflib
import sys
from typing import Iterable, Sequence, Union

from pw_cli.color import colors

_COLOR = colors()


def colorize_diff_line(line: str) -> str:
    if line.startswith('--- ') or line.startswith('+++ '):
        return _COLOR.bold_white(line)
    if line.startswith('-'):
        return _COLOR.red(line)
    if line.startswith('+'):
        return _COLOR.green(line)
    if line.startswith('@@ '):
        return _COLOR.cyan(line)
    return line


def colorize_diff(lines: Union[str, Iterable[str]]) -> str:
    """Takes a diff str or list of str lines and returns a colorized version."""
    if isinstance(lines, str):
        lines = lines.splitlines(True)

    return ''.join(colorize_diff_line(line) for line in lines)


def print_diff(
    a: Sequence[str], b: Sequence[str], fromfile="", tofile="", indent=0
) -> None:
    """Print a git-style diff of two string sequences."""
    indent_str = ' ' * indent
    diff = colorize_diff(difflib.unified_diff(a, b, fromfile, tofile))

    for line in diff.splitlines(True):
        sys.stdout.write(indent_str + line)

    sys.stdout.write("\n")



#!/usr/bin/env python
# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""User API"""

import io
import os
import subprocess
import tempfile

from pathlib import Path
from typing import Any

from pw_emu.core import (
    AlreadyRunning,
    Config,
    ConfigError,
    Connector,
    Launcher,
    InvalidEmulator,
    InvalidChannelType,
    NotRunning,
)


class Emulator:
    """Launches, controls and interacts with an emulator instance."""

    def __init__(self, wdir: Path, config_path: Path | None = None) -> None:
        self._wdir = wdir
        self._config_path = config_path
        self._connector: Connector | None = None
        self._launcher: Launcher | None = None

    def _get_launcher(self, target: str) -> Launcher:
        """Returns an emulator for a given target.

        If there are multiple emulators for the same target it will return
        an arbitrary emulator launcher.
        """
        config = Config(self._config_path)
        target_config = config.get(
            ['targets', target],
            optional=False,
            entry_type=dict,
        )
        for key in target_config.keys():
            try:
                return Launcher.get(key, self._config_path)
            except InvalidEmulator:
                pass
        raise ConfigError(
            self._config_path,
            f'could not determine emulator for target `{target}`',
        )

    def start(
        self,
        target: str,
        file: Path | None = None,
        pause: bool = False,
        debug: bool = False,
        foreground: bool = False,
        args: str | None = None,
    ) -> None:
        """Starts the emulator for the given ``target``.

        If ``file`` is set the emulator loads the file before starting.

        If ``pause`` is ``True`` the emulator pauses until the debugger is
        connected.

        If ``debug`` is ``True`` the emulator runs in the foreground with debug
        output enabled. This is useful for seeing errors, traces, etc.

        If ``foreground`` is ``True`` the emulator runs in the foreground,
        otherwise it starts in daemon mode. Foreground mode is useful when
        there is another process controlling the emulator's life cycle,
        e.g. cuttlefish.

        ``args`` are passed directly to the emulator.

        """
        if self._connector:
            raise AlreadyRunning(self._wdir)

        if self._launcher is None:
            self._launcher = self._get_launcher(target)
        self._connector = self._launcher.start(
            wdir=self._wdir,
            target=target,
            file=file,
            pause=pause,
            debug=debug,
            foreground=foreground,
            args=args,
        )

    def _c(self) -> Connector:
        if self._connector is None:
            self._connector = Connector.get(self._wdir)
            if not self.running():
                raise NotRunning(self._wdir)
        return self._connector

    def running(self) -> bool:
        """Checks if the main emulator process is already running."""

        try:
            return self._c().running()
        except NotRunning:
            return False

    def _path(self, name: Path | str) -> Path | str:
        """Returns the full path for a given emulator file."""

        return os.path.join(self._wdir, name)

    def stop(self):
        """Stop the emulator."""

        return self._c().stop()

    def get_gdb_remote(self) -> str:
        """Returns a string that can be passed to the target remote ``gdb``
        command.

        """

        chan_type = self._c().get_channel_type('gdb')

        if chan_type == 'tcp':
            host, port = self._c().get_channel_addr('gdb')
            return f'{host}:{port}'

        if chan_type == 'pty':
            return self._c().get_channel_path('gdb')

        raise InvalidChannelType(chan_type)

    def get_gdb_cmd(self) -> list[str]:
        """Returns the ``gdb`` command for current target."""
        return self._c().get_gdb_cmd()

    def run_gdb_cmds(
        self,
        commands: list[str],
        executable: str | None = None,
        pause: bool = False,
    ) -> subprocess.CompletedProcess:
        """Connects to the target and runs the given commands silently
        in batch mode.

        ``executable`` is optional but may be required by some ``gdb`` commands.

        If ``pause`` is set, execution stops after running the given commands.

        """

        cmd = self._c().get_gdb_cmd().copy()
        if not cmd:
            raise ConfigError(self._c().get_config_path(), 'gdb not configured')

        cmd.append('-batch-silent')
        cmd.append('-ex')
        cmd.append(f'target remote {self.get_gdb_remote()}')
        for gdb_cmd in commands:
            cmd.append('-ex')
            cmd.append(gdb_cmd)
        if pause:
            cmd.append('-ex')
            cmd.append('disconnect')
        if executable:
            cmd.append(executable)
        return subprocess.run(cmd, capture_output=True)

    def reset(self) -> None:
        """Performs a software reset."""
        self._c().reset()

    def list_properties(self, path: str) -> list[dict]:
        """Returns the property list for an emulator object.

        The object is identified by a full path. The path is
        target-specific and the format of the path is backend-specific.

        QEMU path example: ``/machine/unattached/device[10]``

        renode path example: ``sysbus.uart``

        """
        return self._c().list_properties(path)

    def set_property(self, path: str, prop: str, value: Any) -> None:
        """Sets the value of an emulator's object property."""

        self._c().set_property(path, prop, value)

    def get_property(self, path: str, prop: str) -> Any:
        """Returns the value of an emulator's object property."""

        return self._c().get_property(path, prop)

    def get_channel_type(self, name: str) -> str:
        """Returns the channel type

        Currently ``pty`` and ``tcp`` are the only supported types.

        """

        return self._c().get_channel_type(name)

    def get_channel_path(self, name: str) -> str:
        """Returns the channel path. Raises ``InvalidChannelType`` if this
        is not a ``pty`` channel.

        """

        return self._c().get_channel_path(name)

    def get_channel_addr(self, name: str) -> tuple:
        """Returns a pair of ``(host, port)`` for the channel. Raises
        ``InvalidChannelType`` if this is not a TCP channel.

        """

        return self._c().get_channel_addr(name)

    def get_channel_stream(
        self,
        name: str,
        timeout: float | None = None,
    ) -> io.RawIOBase:
        """Returns a file object for a given host-exposed device.

        If ``timeout`` is ``None`` than reads and writes are blocking. If
        ``timeout`` is ``0`` the stream is operating in non-blocking
        mode. Otherwise read and write will timeout after the given
        value.

        """

        return self._c().get_channel_stream(name, timeout)

    def get_channels(self) -> list[str]:
        """Returns the list of available channels."""

        return self._c().get_channels()

    def set_emu(self, emu: str) -> None:
        """Sets the emulator type for this instance."""

        self._launcher = Launcher.get(emu, self._config_path)

    def cont(self) -> None:
        """Resumes the emulator's execution."""

        self._c().cont()


class TemporaryEmulator(Emulator):
    """Temporary emulator instances.

    Manages emulator instances that run in temporary working
    directories. The emulator instance is stopped and the working
    directory is cleared when the ``with`` block completes.

    It also supports interoperability with the ``pw_emu`` cli, e.g. starting the
    emulator with the CLI and then controlling it from the Python API.

    Usage example:

    .. code-block:: python

        # programatically start and load an executable then access it
        with TemporaryEmulator() as emu:
            emu.start(target, file)
            with emu.get_channel_stream(chan) as stream:
                ...

    .. code-block:: python

        # or start it form the command line then access it
        with TemporaryEmulator() as emu:
            build.bazel(
                ctx,
                "run",
                exec_path,
                "--run_under=pw emu start <target> --file "
            )
            with emu.get_channel_stream(chan) as stream:
                ...

    """

    def __init__(
        self,
        config_path: Path | None = None,
        cleanup: bool = True,
    ) -> None:
        self._temp = tempfile.TemporaryDirectory(prefix="pw_emu_")
        self._cleanup = cleanup
        super().__init__(Path(self._temp.name), config_path)

    def __enter__(self):
        # Interoperability with pw emu cli.
        os.environ["PW_EMU_WDIR"] = str(self._wdir)
        return self

    def __exit__(self, exc, value, traceback) -> None:
        self.stop()
        del os.environ["PW_EMU_WDIR"]
        if self._cleanup:
            self._temp.cleanup()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_tokenizer_proto/options.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n pw_tokenizer_proto/options.proto\x12\x0cpw.tokenizer\x1a google/protobuf/descriptor.proto*)\n\x0cTokenization\x12\x19\n\x15TOKENIZATION_OPTIONAL\x10\x00:K\n\x06\x66ormat\x12\x1d.google.protobuf.FieldOptions\x18\xf0\xe5\x04 \x01(\x0e\x32\x1a.pw.tokenizer.Tokenizationb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_tokenizer_proto.options_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_TOKENIZATION']._serialized_start=84
  _globals['_TOKENIZATION']._serialized_end=125
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_cpu_exception_risc_v_protos/cpu_state.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.pw_cpu_exception_risc_v_protos/cpu_state.proto\x12\x17pw.cpu_exception.risc_v\"\xab\x03\n\rRiscvCpuState\x12\x0c\n\x04mepc\x18\x01 \x01(\r\x12\x0e\n\x06mcause\x18\x02 \x01(\r\x12\x0f\n\x07mstatus\x18\x03 \x01(\r\x12\r\n\x05mtval\x18\x04 \x01(\r\x12\n\n\x02ra\x18\x05 \x01(\r\x12\n\n\x02sp\x18\x06 \x01(\r\x12\n\n\x02t0\x18\x07 \x01(\r\x12\n\n\x02t1\x18\x08 \x01(\r\x12\n\n\x02t2\x18\t \x01(\r\x12\n\n\x02\x66p\x18\n \x01(\r\x12\n\n\x02s1\x18\x0b \x01(\r\x12\n\n\x02\x61\x30\x18\x0c \x01(\r\x12\n\n\x02\x61\x31\x18\r \x01(\r\x12\n\n\x02\x61\x32\x18\x0e \x01(\r\x12\n\n\x02\x61\x33\x18\x0f \x01(\r\x12\n\n\x02\x61\x34\x18\x10 \x01(\r\x12\n\n\x02\x61\x35\x18\x11 \x01(\r\x12\n\n\x02\x61\x36\x18\x12 \x01(\r\x12\n\n\x02\x61\x37\x18\x1d \x01(\r\x12\n\n\x02s2\x18\x14 \x01(\r\x12\n\n\x02s3\x18\x15 \x01(\r\x12\n\n\x02s4\x18\x16 \x01(\r\x12\n\n\x02s5\x18\x17 \x01(\r\x12\n\n\x02s6\x18\x18 \x01(\r\x12\n\n\x02s7\x18\x19 \x01(\r\x12\n\n\x02s8\x18\x1a \x01(\r\x12\n\n\x02s9\x18\x1b \x01(\r\x12\x0b\n\x03s10\x18\x1c \x01(\r\x12\x0b\n\x03s11\x18\' \x01(\r\x12\n\n\x02t3\x18\x1e \x01(\r\x12\n\n\x02t4\x18\x1f \x01(\r\x12\n\n\x02t5\x18  \x01(\r\x12\n\n\x02t6\x18! \x01(\r\"Z\n\x17SnapshotCpuStateOverlay\x12?\n\x0friscv_cpu_state\x18\x14 \x01(\x0b\x32&.pw.cpu_exception.risc_v.RiscvCpuState')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_cpu_exception_risc_v_protos.cpu_state_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_RISCVCPUSTATE']._serialized_start=76
  _globals['_RISCVCPUSTATE']._serialized_end=503
  _globals['_SNAPSHOTCPUSTATEOVERLAY']._serialized_start=505
  _globals['_SNAPSHOTCPUSTATEOVERLAY']._serialized_end=595
# @@protoc_insertion_point(module_scope)



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Launch a simulated device subprocess under pw_console."""

import argparse
from pathlib import Path
import subprocess
import sys
from types import ModuleType

from pw_system.console import console, get_parser
from pw_system.device_connection import DeviceConnection


def _parse_args() -> tuple[argparse.Namespace, list[str]]:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--sim-binary',
        type=Path,
        help='Path to a simulated device binary to run on the host',
    )
    return parser.parse_known_args()


def launch_sim(
    sim_binary: Path,
    remaining_args: list[str],
    compiled_protos: list[ModuleType] | None = None,
    device_connection: DeviceConnection | None = None,
) -> int:
    """Launches a host-device-sim binary, and attaches a console to it."""
    if '-s' not in remaining_args and '--socket-addr' not in remaining_args:
        remaining_args.extend(['--socket-addr', 'default'])

    # Pass the remaining arguments on to pw_system console.
    console_args = get_parser().parse_args(remaining_args)

    sim_process = subprocess.Popen(
        sim_binary,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.STDOUT,
    )

    try:
        retval = console(
            **vars(console_args),
            compiled_protos=compiled_protos,
            device_connection=device_connection,
        )
    finally:
        sim_process.terminate()

    return retval


def main(
    compiled_protos: list[ModuleType] | None = None,
    device_connection: DeviceConnection | None = None,
) -> int:
    sim_args, remaining_args = _parse_args()
    return launch_sim(
        **vars(sim_args),
        remaining_args=remaining_args,
        compiled_protos=compiled_protos,
        device_connection=device_connection,
    )


# TODO(b/353327855): Deprecated function, remove when no longer used.
def main_with_compiled_protos(
    compiled_protos: list[ModuleType] | None = None,
    device_connection: DeviceConnection | None = None,
) -> int:
    return main(
        compiled_protos=compiled_protos,
        device_connection=device_connection,
    )


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Manages SEED documents in Pigweed."""

import argparse
import datetime
import enum
from dataclasses import dataclass
from pathlib import Path
import random
import re
import subprocess
import sys
import urllib.request
from typing import Dict, Iterable, List, Optional, Tuple, Union

import pw_cli.color
import pw_cli.env
from pw_cli.git_repo import GitRepo
from pw_cli.tool_runner import BasicSubprocessRunner

_NEW_SEED_TEMPLATE = '''.. _seed-{num:04d}:

{title_underline}
{formatted_title}
{title_underline}
.. seed::
   :number: {num}
   :name: {title}
   :status: {status}
   :proposal_date: {date}
   :cl: {changelist}
   :authors: {authors}
   :facilitator: Unassigned

-------
Summary
-------
Write up your proposal here.
'''


class SeedStatus(enum.Enum):
    """Possible states of a SEED proposal."""

    DRAFT = 0
    OPEN_FOR_COMMENTS = 1
    LAST_CALL = 2
    ACCEPTED = 3
    REJECTED = 4
    DEPRECATED = 5
    SUPERSEDED = 6
    ON_HOLD = 7
    META = 8

    def __str__(self) -> str:
        if self is SeedStatus.DRAFT:
            return 'Draft'
        if self is SeedStatus.OPEN_FOR_COMMENTS:
            return 'Open for Comments'
        if self is SeedStatus.LAST_CALL:
            return 'Last Call'
        if self is SeedStatus.ACCEPTED:
            return 'Accepted'
        if self is SeedStatus.REJECTED:
            return 'Rejected'
        if self is SeedStatus.DEPRECATED:
            return 'Deprecated'
        if self is SeedStatus.SUPERSEDED:
            return 'Superseded'
        if self is SeedStatus.ON_HOLD:
            return 'On Hold'
        if self is SeedStatus.META:
            return 'Meta'

        return ''


@dataclass
class SeedMetadata:
    number: int
    title: str
    authors: str
    status: SeedStatus
    changelist: Optional[int] = None
    sources: Optional[List[str]] = None

    def default_filename(self) -> str:
        normalized_title = self.title.lower().replace(' ', '-')
        normalized_title = re.sub(r'[^a-zA-Z0-9_-]', '', normalized_title)

        return f'{self.number:04d}-{normalized_title}.rst'


class SeedRegistry:
    """
    Represents a registry of SEEDs located somewhere in pigweed.git, which can
    be read, modified, and written.

    Currently, this is implemented as a basic text parser for the BUILD.gn file
    in the seed/ directory; however, in the future it may be rewritten to use a
    different backing data source.
    """

    class _State(enum.Enum):
        OUTER = 0
        SEED = 1
        INDEX = 2
        INDEX_SEEDS_LIST = 3

    @classmethod
    def parse(cls, registry_file: Path) -> 'SeedRegistry':
        return cls(registry_file)

    def __init__(self, seed_build_file: Path):
        self._file = seed_build_file
        self._lines = seed_build_file.read_text().split('\n')
        self._seeds: Dict[int, Tuple[int, int]] = {}
        self._next_seed_number = 101

        seed_regex = re.compile(r'pw_seed\("(\d+)"\)')

        state = SeedRegistry._State.OUTER

        section_start_index = 0
        current_seed_number = 0

        # Run through the GN file, doing some basic parsing of its targets.
        for i, line in enumerate(self._lines):
            if state is SeedRegistry._State.OUTER:
                seed_match = seed_regex.search(line)
                if seed_match:
                    # SEED definition target: extract the number.
                    state = SeedRegistry._State.SEED

                    section_start_index = i
                    current_seed_number = int(seed_match.group(1))

                    if current_seed_number >= self._next_seed_number:
                        self._next_seed_number = current_seed_number + 1

                if line == 'pw_seed_index("seeds") {':
                    state = SeedRegistry._State.INDEX
                    # Skip back past the comments preceding the SEED index
                    # target. New SEEDs will be inserted here.
                    insertion_index = i
                    while insertion_index > 0 and self._lines[
                        insertion_index - 1
                    ].startswith('#'):
                        insertion_index -= 1

                    self._seed_insertion_index = insertion_index

            if state is SeedRegistry._State.SEED:
                if line == '}':
                    self._seeds[current_seed_number] = (section_start_index, i)
                    state = SeedRegistry._State.OUTER

            if state is SeedRegistry._State.INDEX:
                if line == '}':
                    state = SeedRegistry._State.OUTER
                if line == '  seeds = [':
                    state = SeedRegistry._State.INDEX_SEEDS_LIST

            if state is SeedRegistry._State.INDEX_SEEDS_LIST:
                if line == '  ]':
                    self._index_seeds_end = i
                    state = SeedRegistry._State.INDEX

    def file(self) -> Path:
        """Returns the file which backs this registry."""
        return self._file

    def seed_count(self) -> int:
        """Returns the number of SEEDs registered."""
        return len(self._seeds)

    def insert(self, seed: SeedMetadata) -> None:
        """Adds a new seed to the registry."""

        new_seed = [
            f'pw_seed("{seed.number:04d}") {{',
            f'  title = "{seed.title}"',
            f'  author = "{seed.authors}"',
            f'  status = "{seed.status}"',
        ]

        if seed.changelist is not None:
            new_seed.append(f'  changelist = {seed.changelist}')

        if seed.sources is not None:
            if len(seed.sources) == 0:
                new_seed.append('  sources = []')
            elif len(seed.sources) == 1:
                new_seed.append(f'  sources = [ "{seed.sources[0]}" ]')
            else:
                new_seed.append('  sources = [')
                new_seed.extend(f'    "{source}",' for source in seed.sources)
                new_seed.append('  ]')

        new_seed += [
            '}',
            '',
        ]
        self._lines = (
            self._lines[: self._seed_insertion_index]
            + new_seed
            + self._lines[self._seed_insertion_index : self._index_seeds_end]
            + [f'    ":{seed.number:04d}",']
            + self._lines[self._index_seeds_end :]
        )

        self._seed_insertion_index += len(new_seed)
        self._index_seeds_end += len(new_seed)

        if seed.number == self._next_seed_number:
            self._next_seed_number += 1

    def next_seed_number(self) -> int:
        return self._next_seed_number

    def write(self) -> None:
        self._file.write_text('\n'.join(self._lines))


_GERRIT_HOOK_URL = (
    'https://gerrit-review.googlesource.com/tools/hooks/commit-msg'
)


# TODO: pwbug.dev/318746837 - Extract this to somewhere more general.
def _install_gerrit_hook(git_root: Path) -> None:
    hook_file = git_root / '.git' / 'hooks' / 'commit-msg'
    urllib.request.urlretrieve(_GERRIT_HOOK_URL, hook_file)
    hook_file.chmod(0o755)


def _request_new_seed_metadata(
    repo: GitRepo,
    registry: SeedRegistry,
    colors,
) -> SeedMetadata:
    if repo.has_uncommitted_changes():
        print(
            colors.red(
                'You have uncommitted Git changes. '
                'Please commit or stash before creating a SEED.'
            )
        )
        sys.exit(1)

    print(
        colors.yellow(
            'This command will create Git commits. '
            'Make sure you are on a suitable branch.'
        )
    )
    print(f'Current branch: {colors.cyan(repo.current_branch())}')
    print('')

    number = registry.next_seed_number()

    try:
        num = input(
            f'SEED number (default={colors.bold_white(number)}): '
        ).strip()

        while True:
            try:
                if num:
                    number = int(num)
                break
            except ValueError:
                num = input('Invalid number entered. Try again: ').strip()

        title = input('SEED title: ').strip()
        while not title:
            title = input(
                'Title cannot be empty. Re-enter SEED title: '
            ).strip()

        authors = input('SEED authors: ').strip()
        while not authors:
            authors = input(
                'Authors list cannot be empty. Re-enter SEED authors: '
            ).strip()

        print('The following SEED will be created.')
        print('')
        print(f'  Number: {colors.green(number)}')
        print(f'  Title: {colors.green(title)}')
        print(f'  Authors: {colors.green(authors)}')
        print('')
        print(
            'This will create two commits on branch '
            + colors.cyan(repo.current_branch())
            + ' and push them to Gerrit.'
        )

        create = True
        confirm = input(f'Proceed? [{colors.bold_white("Y")}/n] ').strip()
        if confirm:
            create = confirm == 'Y'

    except KeyboardInterrupt:
        print('\nReceived CTRL-C, exiting...')
        sys.exit(0)

    if not create:
        sys.exit(0)

    return SeedMetadata(
        number=number,
        title=title,
        authors=authors,
        status=SeedStatus.DRAFT,
    )


@dataclass
class GerritChange:
    id: str
    number: int
    title: str

    def url(self) -> str:
        return (
            'https://pigweed-review.googlesource.com'
            f'/c/pigweed/pigweed/+/{self.number}'
        )


def commit_and_push(
    repo: GitRepo,
    files: Iterable[Union[Path, str]],
    commit_message: str,
    change_id: Optional[str] = None,
) -> GerritChange:
    """Creates a commit with the given files and message and pushes to Gerrit.

    Args:
        change_id: Optional Gerrit change ID to use. If not specified, generates
            a new one.
    """
    if change_id is not None:
        commit_message = f'{commit_message}\n\nChange-Id: {change_id}'

    subprocess.run(
        ['git', 'add'] + list(files), capture_output=True, check=True
    )
    subprocess.run(
        ['git', 'commit', '-m', commit_message], capture_output=True, check=True
    )

    if change_id is None:
        # Parse the generated change ID from the commit if it wasn't
        # explicitly set.
        change_id = repo.commit_change_id()

        if change_id is None:
            # If the commit doesn't have a Change-Id, the Gerrit hook is not
            # installed. Install it and try modifying the commit.
            _install_gerrit_hook(repo.root())
            subprocess.run(
                ['git', 'commit', '--amend', '--no-edit'],
                capture_output=True,
                check=True,
            )
            change_id = repo.commit_change_id()
            assert change_id is not None

    process = subprocess.run(
        [
            'git',
            'push',
            'origin',
            '+HEAD:refs/for/main',
            '--no-verify',
        ],
        capture_output=True,
        text=True,
        check=True,
    )

    output = process.stderr

    regex = re.compile(
        '^\\s*remote:\\s*'
        'https://pigweed-review.(?:git.corp.google|googlesource).com/'
        'c/pigweed/pigweed/\\+/(?P<num>\\d+)\\s+',
        re.MULTILINE,
    )
    match = regex.search(output)
    if not match:
        raise ValueError(f"invalid output from 'git push': {output}")
    change_num = int(match.group('num'))

    return GerritChange(change_id, change_num, commit_message.split('\n')[0])


def _create_wip_seed_doc_change(
    repo: GitRepo,
    new_seed: SeedMetadata,
) -> GerritChange:
    """Commits and pushes a boilerplate CL doc to Gerrit.

    Returns information about the CL.
    """
    env = pw_cli.env.pigweed_environment()
    seed_rst_file = env.PW_ROOT / 'seed' / new_seed.default_filename()

    formatted_title = f'{new_seed.number:04d}: {new_seed.title}'
    title_underline = '=' * len(formatted_title)

    seed_rst_file.write_text(
        _NEW_SEED_TEMPLATE.format(
            formatted_title=formatted_title,
            title_underline=title_underline,
            num=new_seed.number,
            title=new_seed.title,
            authors=new_seed.authors,
            status=new_seed.status,
            date=datetime.date.today().strftime('%Y-%m-%d'),
            changelist=0,
        )
    )

    temp_branch = f'wip-seed-{new_seed.number}-{random.randrange(0, 2**16):x}'
    subprocess.run(
        ['git', 'checkout', '-b', temp_branch], capture_output=True, check=True
    )

    commit_message = f'SEED-{new_seed.number:04d}: {new_seed.title}'

    try:
        cl = commit_and_push(repo, [seed_rst_file], commit_message)
    except subprocess.CalledProcessError as err:
        print(f'Command {err.cmd} failed; stderr:')
        print(err.stderr)
        sys.exit(1)
    finally:
        subprocess.run(
            ['git', 'checkout', '-'], capture_output=True, check=True
        )
        subprocess.run(
            ['git', 'branch', '-D', temp_branch],
            capture_output=True,
            check=True,
        )

    return cl


def create_seed_number_claim_change(
    repo: GitRepo,
    new_seed: SeedMetadata,
    registry: SeedRegistry,
) -> GerritChange:
    commit_message = f'SEED-{new_seed.number:04d}: Claim SEED number'
    registry.insert(new_seed)
    registry.write()
    return commit_and_push(repo, [registry.file()], commit_message)


def _create_seed_doc_change(
    repo: GitRepo,
    new_seed: SeedMetadata,
    registry: SeedRegistry,
    wip_change: GerritChange,
) -> None:
    env = pw_cli.env.pigweed_environment()
    seed_rst_file = env.PW_ROOT / 'seed' / new_seed.default_filename()

    formatted_title = f'{new_seed.number:04d}: {new_seed.title}'
    title_underline = '=' * len(formatted_title)

    seed_rst_file.write_text(
        _NEW_SEED_TEMPLATE.format(
            formatted_title=formatted_title,
            title_underline=title_underline,
            num=new_seed.number,
            title=new_seed.title,
            authors=new_seed.authors,
            status=new_seed.status,
            date=datetime.date.today().strftime('%Y-%m-%d'),
            changelist=new_seed.changelist,
        )
    )

    new_seed.sources = [seed_rst_file.relative_to(registry.file().parent)]
    new_seed.changelist = None
    registry.insert(new_seed)
    registry.write()

    commit_message = f'SEED-{new_seed.number:04d}: {new_seed.title}'
    commit_and_push(
        repo,
        [registry.file(), seed_rst_file],
        commit_message,
        change_id=wip_change.id,
    )


def create_seed() -> int:
    colors = pw_cli.color.colors()
    env = pw_cli.env.pigweed_environment()

    repo = GitRepo(env.PW_ROOT, BasicSubprocessRunner())

    registry_path = env.PW_ROOT / 'seed' / 'BUILD.gn'

    wip_registry = SeedRegistry.parse(registry_path)
    registry = SeedRegistry.parse(registry_path)

    seed = _request_new_seed_metadata(repo, wip_registry, colors)
    seed_cl = _create_wip_seed_doc_change(repo, seed)

    seed.changelist = seed_cl.number

    number_cl = create_seed_number_claim_change(repo, seed, wip_registry)
    _create_seed_doc_change(repo, seed, registry, seed_cl)

    print()
    print(f'Created two CLs for SEED-{seed.number:04d}:')
    print()
    print(f'-  {number_cl.title}')
    print(f'   <{number_cl.url()}>')
    print()
    print(f'-  {seed_cl.title}')
    print(f'   <{seed_cl.url()}>')

    return 0


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.set_defaults(func=lambda **_kwargs: parser.print_help())

    subparsers = parser.add_subparsers(title='subcommands')

    # No args for now, as this initially only runs in interactive mode.
    create_parser = subparsers.add_parser('create', help='Creates a new SEED')
    create_parser.set_defaults(func=create_seed)

    return parser.parse_args()


def main() -> int:
    args = {**vars(parse_args())}
    func = args['func']
    del args['func']

    exit_code = func(**args)
    return 0 if exit_code is None else exit_code


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Find serial ports."""

import argparse
import logging
import operator
import sys
from typing import Optional, Sequence

from serial.tools.list_ports import comports
from serial.tools.list_ports_common import ListPortInfo

from pw_cli.interactive_prompts import interactive_index_select

_LOG = logging.getLogger(__package__)


def _parse_args():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '-i',
        '--interactive',
        action='store_true',
        help='Show an interactive prompt to select a port.',
    )
    parser.add_argument(
        '-l',
        '--list-ports',
        action='store_true',
        help='List all port info.',
    )
    parser.add_argument(
        '-p',
        '--product',
        help='Print ports matching this product name.',
    )
    parser.add_argument(
        '-m',
        '--manufacturer',
        help='Print ports matching this manufacturer name.',
    )
    parser.add_argument(
        '-s',
        '--serial-number',
        help='Print ports matching this serial number.',
    )
    parser.add_argument(
        '-1',
        '--print-first-match',
        action='store_true',
        help='Print the first port found sorted by device path.',
    )
    return parser.parse_args()


def _print_ports(ports: Sequence[ListPortInfo]):
    for cp in ports:
        for line in [
            f"device        = {cp.device}",
            f"name          = {cp.name}",
            f"description   = {cp.description}",
            f"vid           = {cp.vid}",
            f"pid           = {cp.pid}",
            f"serial_number = {cp.serial_number}",
            f"location      = {cp.location}",
            f"manufacturer  = {cp.manufacturer}",
            f"product       = {cp.product}",
            f"interface     = {cp.interface}",
        ]:
            print(line)
        print()


def main(
    list_ports: bool = False,
    product: Optional[str] = None,
    manufacturer: Optional[str] = None,
    serial_number: Optional[str] = None,
    print_first_match: bool = False,
    interactive: bool = False,
) -> int:
    """List device info or print matches."""
    ports = sorted(comports(), key=operator.attrgetter('device'))

    if list_ports:
        _print_ports(ports)
        return 0

    if interactive:
        selected_port = interactive_serial_port_select()
        if selected_port:
            print(selected_port)
            return 0
        return 1

    any_match_found = False

    # Print matching devices
    for port in ports:
        if (
            product is not None
            and port.product is not None
            and product in port.product
        ):
            any_match_found = True
            print(port.device)

        if (
            manufacturer is not None
            and port.manufacturer is not None
            and manufacturer in port.manufacturer
        ):
            any_match_found = True
            print(port.device)

        if (
            serial_number is not None
            and port.serial_number is not None
            and serial_number in port.serial_number
        ):
            any_match_found = True
            print(port.device)

        if any_match_found and print_first_match:
            return 0

    if not any_match_found:
        return 1

    return 0


def interactive_serial_port_select(auto_select_only_one=True) -> str:
    """Prompt the user to select a detected serial port.

    Returns: String containing the path to the tty device.
    """
    ports = sorted(comports(), key=operator.attrgetter('device'))

    if not ports:
        print()
        print(
            '\033[31mERROR:\033[0m No serial ports detected.',
            file=sys.stderr,
        )
        sys.exit(1)

    if auto_select_only_one and len(ports) == 1:
        # Auto select the first port.
        return ports[0].device

    # Create valid entry list
    port_lines = list(
        f'{port.device} - {port.manufacturer} - {port.description}'
        for i, port in enumerate(ports)
    )

    print()
    print('Please select a serial port device.')
    print('Available ports:')
    selected_index, _selected_text = interactive_index_select(
        selection_lines=port_lines,
        prompt_text=(
            'Enter a port index or press up/down (Ctrl-C to cancel)\n> '
        ),
    )

    selected_port = ports[selected_index]

    return selected_port.device


if __name__ == '__main__':
    sys.exit(main(**vars(_parse_args())))



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Generate a Python wheel for a pw_python_package or pw_python_distribution."""

import argparse
import hashlib
from pathlib import Path
import shutil
import subprocess
import sys

try:
    from pw_build.python_package import PythonPackage

except ImportError:
    # Load from python_package from this directory if pw_build is not available.
    from python_package import PythonPackage  # type: ignore


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--package-dir',
        type=Path,
        required=True,
        help='Path to the root gn build dir.',
    )
    parser.add_argument(
        '--out-dir',
        type=Path,
        required=True,
        help='requirement file to generate',
    )
    parser.add_argument(
        '--generate-hashes',
        action='store_true',
        help='Generate sha256 sums for the requirements.txt file.',
    )
    return parser.parse_args()


def main(
    package_dir: Path,
    out_dir: Path,
    generate_hashes: bool,
) -> int:
    """Build a Python wheel."""

    # Delete existing wheels from the out dir, there may be stale versions.
    shutil.rmtree(out_dir, ignore_errors=True)
    out_dir.mkdir(parents=True, exist_ok=True)

    # Find the target package name and version.
    pkg = PythonPackage(
        sources=[],
        setup_sources=[package_dir / 'setup.cfg'],
        tests=[],
        inputs=[],
    )
    requirement_lines = f'{pkg.package_name}=={pkg.package_version}'

    # Build the wheel.
    subprocess.run(
        [
            sys.executable,
            "-m",
            "build",
            str(package_dir),
            "--wheel",
            "--no-isolation",
            "--outdir",
            str(out_dir),
        ],
        check=True,
    )

    if generate_hashes:
        # Cacluate the sha256
        for wheel_path in out_dir.glob('**/*.whl'):
            wheel_sha256 = hashlib.sha256()
            wheel_sha256.update(wheel_path.read_bytes())
            sha256 = wheel_sha256.hexdigest()
            requirement_lines += ' \\\n    '
            requirement_lines += f'--hash=sha256:{sha256}'
            break

    # Save a requirements file for this wheel and hash value.
    requirement_file = out_dir / 'requirements.txt'
    requirement_file.write_text(requirement_lines, encoding='utf-8')

    return 0


if __name__ == '__main__':
    sys.exit(main(**vars(_parse_args())))
